"""
Set of syntactic sugar processors at the intermediate level.
"""

import operator as op
from typing import AbstractSet, Callable, DefaultDict

from ... import expression_walker as ew, expressions as ir
from ...datalog.expression_processing import (
    conjunct_formulas,
    extract_logic_atoms,
)
from ...exceptions import SymbolNotFoundError
from ...expression_walker import ReplaceExpressionWalker
from ...expressions import Constant, FunctionApplication, Symbol
from ...logic import TRUE, Conjunction, Implication
from ...type_system import Unknown, get_args, is_leq_informative


class Column(ir.Definition):
    def __init__(self, set_symbol, column_position):
        self.set_symbol = set_symbol
        self.column_position = column_position
        self._symbols = (
            self.set_symbol._symbols | self.column_position._symbols
        )


def _has_column_sugar(conjunction, marker):
    return any(
        isinstance(arg, marker)
        for atom in extract_logic_atoms(conjunction)
        for arg in atom.args
    )


class TranslateColumnsToAtoms(ew.PatternWalker):
    """
    Syntactic sugar to handle cases where the first column is used
    as a selector. Specifically cases such as

    >>> Implication(B(z), C(z, Column(A, 1)))

    is transformed to

    >>> Implication(B(z), Conjunction((A(fresh0, fresh1), C(z, fresh1))))

    If this syntactic sugar is on the head, then every atom that, by type
    has two arguments, but only one is used, will be replaced by the case
    where first argument is the sugared element.

    >>> Implication(Column(A, 2), B(Column(A, 2), x)))

    is transformed to

    >>> Implication(A(fresh0, fresh1, fresh2), B(fresh2, x)))
    """

    @ew.add_match(
        ir.FunctionApplication,
        lambda exp: any(isinstance(arg, Column) for arg in exp.args),
    )
    def function_application_column_sugar(self, expression):
        return self.walk(Conjunction((expression,)))

    @ew.add_match(Conjunction, lambda exp: _has_column_sugar(exp, Column))
    def conjunction_column_sugar(self, expression):
        (
            replacements,
            new_atoms,
        ) = self._obtain_new_atoms_and_column_replacements(expression)

        replaced_expression = ew.ReplaceExpressionWalker(replacements).walk(
            expression
        )
        new_formulas = replaced_expression.formulas + tuple(new_atoms)
        return self.walk(Conjunction(new_formulas))

    @ew.add_match(Implication, lambda exp: _has_column_sugar(exp, Column))
    def implication_column_sugar(self, expression):
        (
            replacements,
            new_atoms,
        ) = self._obtain_new_atoms_and_column_replacements(expression)

        replaced_expression = ew.ReplaceExpressionWalker(replacements).walk(
            expression
        )
        new_antecedent = conjunct_formulas(
            replaced_expression.antecedent, Conjunction(new_atoms)
        )

        return self.walk(
            Implication(replaced_expression.consequent, new_antecedent)
        )

    def _obtain_new_atoms_and_column_replacements(self, expression):
        sugared_columns = self._obtain_sugared_columns(expression)

        new_atoms = []
        replacements = {}
        for k, v in sugared_columns.items():
            k_constant = ew.ReplaceSymbolsByConstants(self.symbol_table).walk(
                k
            )
            args = (
                v.get(Column(k, ir.Constant[int](i)), ir.Symbol.fresh())
                for i in range(k_constant.value.arity)
            )
            new_atoms.append(k(*args))
            replacements.update(v)
        return replacements, tuple(new_atoms)

    def _obtain_sugared_columns(self, expression):
        sugared_columns = DefaultDict(dict)
        for atom in extract_logic_atoms(expression):
            for arg in atom.args:
                if isinstance(arg, Column):
                    sugared_columns[arg.set_symbol][arg] = ir.Symbol.fresh()
        return sugared_columns


class SelectByFirstColumn(ir.Definition):
    def __init__(self, set_symbol, selector):
        self.set_symbol = set_symbol
        self.selector = selector
        self._symbols = self.set_symbol._symbols | self.selector._symbols

    def __repr__(self):
        return f"{self.set_symbol}.{self.selector}"


class TranslateSelectByFirstColumn(ew.PatternWalker):
    """
    Syntactic sugar to handle cases where the first column is used
    as a selector. Specifically cases such as

    >>> Implication(B(z), C(z, SelectByFirstColumn(A, c)))

    is transformed to

    >>> Implication(B(z), Conjunction((A(c, fresh), C(z, fresh))))

    If this syntactic sugar is on the head, then every atom that, by type
    has two arguments, but only one is used, will be replaced by the case
    where first argument is the sugared element.

    >>> Implication(SelectByFirstColumn(A, c), Conjunction((eq(x), B(x))))

    is transformed to

    >>> Implication(A(c, fresh), Conjunction((eq(fresh, x), B(x))))
    """

    @ew.add_match(
        ir.FunctionApplication,
        lambda exp: any(
            isinstance(arg, SelectByFirstColumn) for arg in exp.args
        ),
    )
    def function_application_column_sugar(self, expression):
        return self.walk(Conjunction((expression,)))

    @ew.add_match(
        Conjunction, lambda exp: _has_column_sugar(exp, SelectByFirstColumn)
    )
    def conjunction_column_sugar(self, expression):
        (
            replacements,
            new_atoms,
        ) = self._obtain_new_atoms_and_select_by_first_column_replacements(
            expression
        )

        replaced_expression = ew.ReplaceExpressionWalker(replacements).walk(
            expression
        )
        new_formulas = replaced_expression.formulas + tuple(new_atoms)
        return self.walk(Conjunction(new_formulas))

    @ew.add_match(Implication(SelectByFirstColumn, ...))
    def implication_select_by_left_head(self, expression):
        consequent = expression.consequent
        head_fresh = ir.Symbol.fresh()
        new_consequent = consequent.set_symbol(consequent.selector, head_fresh)

        replacements = {consequent: new_consequent}
        for atom in extract_logic_atoms(expression.antecedent):
            if len(atom.args) == 1 and self._theoretical_arity(atom) == 2:
                replacements[atom] = atom.functor(head_fresh, atom.args[0])

        new_rule = Implication(
            new_consequent,
            ReplaceExpressionWalker(replacements).walk(expression.antecedent),
        )

        return self.walk(new_rule)

    @ew.add_match(
        Implication, lambda exp: _has_column_sugar(exp, SelectByFirstColumn)
    )
    def implication_column_sugar(self, expression):
        (
            replacements,
            new_atoms,
        ) = self._obtain_new_atoms_and_select_by_first_column_replacements(
            expression
        )

        replaced_expression = ew.ReplaceExpressionWalker(replacements).walk(
            expression
        )
        new_antecedent = conjunct_formulas(
            replaced_expression.antecedent, Conjunction(new_atoms)
        )

        return self.walk(
            Implication(replaced_expression.consequent, new_antecedent)
        )

    def _obtain_new_atoms_and_select_by_first_column_replacements(
        self, expression
    ):
        sugared_columns = DefaultDict(dict)
        for atom in extract_logic_atoms(expression):
            for arg in atom.args:
                if isinstance(arg, SelectByFirstColumn):
                    sugared_columns[arg] = ir.Symbol.fresh()

        new_atoms = []
        for k, v in sugared_columns.items():
            new_atom = k.set_symbol(k.selector, v)
            new_atoms.append(new_atom)
        return sugared_columns, tuple(new_atoms)

    def _theoretical_arity(self, atom):
        functor = atom.functor
        if isinstance(functor, ir.Symbol) and functor.type is Unknown:
            try:
                functor = self.symbol_table[functor]
            except KeyError:
                raise SymbolNotFoundError(f"Symbol {functor} not found")
        if functor.type is Unknown:
            arity = None
        elif is_leq_informative(functor.type, Callable):
            arity = len(get_args(functor.type)[:-1])
        elif is_leq_informative(functor.type, AbstractSet):
            arity = len(get_args(get_args(functor.type)[0]))
        else:
            arity = None
        return arity


GETATTR = ir.Constant(getattr)


class ConvertAttrSToSelectByColumn(ew.ExpressionWalker):
    """
    Convert terms such as P.s[c]
    or `getattr(P, s)[c]` at `SelectByFirstColumn(P, s)`.
    """

    @ew.add_match(
        FunctionApplication(
            FunctionApplication(GETATTR, (..., Constant[str]("s"))), (...,)
        )
    )
    def conversion(self, expression):
        return self.walk(
            SelectByFirstColumn(expression.functor.args[0], expression.args[0])
        )


class RecogniseSSugar(ew.PatternWalker):
    """
    Recognising datalog terms such as P.s[c]
    or `getattr(P, s)[c]`.
    """

    @ew.add_match(Constant)
    def constant(self, expression):
        return False

    @ew.add_match(Symbol)
    def symbol(self, expression):
        return False

    @ew.add_match(
        FunctionApplication(
            FunctionApplication(GETATTR, (..., Constant[str]("s"))), (...,)
        )
    )
    def s_sugar(self, expression):
        return True

    @ew.add_match(...)
    def others(self, expression):
        params = list(expression.unapply())
        while params:
            param = params.pop()
            if isinstance(param, tuple):
                params += param
            else:
                if self.walk(param):
                    return True
        return False


_RECOGNISE_S_SUGAR = RecogniseSSugar()


class TranslateSSugarToSelectByColumn(ew.PatternWalker):
    """
    Syntactic sugar to convert datalog terms P.s[c] to
    SelectByFirstColumn(P, s).
    """

    _convert_attr_s__to_SelectByColumn = ConvertAttrSToSelectByColumn()

    @ew.add_match(Implication, _RECOGNISE_S_SUGAR.walk)
    def replace_s_getattr_by_first_column(self, expression):
        new_expression = (
            TranslateSSugarToSelectByColumn
            ._convert_attr_s__to_SelectByColumn
            .walk(
                expression
            )
        )
        if new_expression is not expression:
            new_expression = self.walk(new_expression)
        return new_expression


EQ = Constant(op.eq)


class TranslateHeadConstantsToEqualities(ew.PatternWalker):
    """
    Syntactic sugar to convert datalog rules having constants
    in the head to having equalities in the body.
    """

    @ew.add_match(
        Implication,
        lambda imp: any(
            isinstance(arg, ir.Constant) for arg in imp.consequent.args
        )
        and (imp.antecedent != TRUE),
    )
    def head_constants_to_equalities(self, expression):
        new_equalities = {}
        new_args = tuple()
        for arg in expression.consequent.args:
            if isinstance(arg, ir.Constant):
                new_param = ir.Symbol.fresh()
                new_equalities[new_param] = arg
                arg = new_param
            new_args += (arg,)
        new_consequent = expression.consequent.functor(*new_args)
        new_antecedent = Conjunction(
            tuple(
                [expression.antecedent]
                + [EQ(k, v) for k, v in new_equalities.items()]
            )
        )

        new_implication = Implication(new_consequent, new_antecedent)

        return self.walk(new_implication)



from ..datalog.basic_representation import (
    DatalogProgram,
    UnionOfConjunctiveQueries,
)
from ..exceptions import ForbiddenExpressionError
from ..expression_pattern_matching import add_match
from ..expression_walker import ExpressionWalker, PatternWalker
from ..expressions import FunctionApplication, Symbol
from ..logic import Implication, Union


def is_ppdl_rule(exp):
    return (
        isinstance(exp, Implication)
        and isinstance(exp.consequent, FunctionApplication)
        and sum(isinstance(arg, PPDLDeltaTerm) for arg in exp.consequent.args)
        == 1
    )


def get_dterm(datom):
    return next(arg for arg in datom.args if isinstance(arg, PPDLDeltaTerm))


def get_dterm_index(datom):
    return next(
        i for i, arg in enumerate(datom.args) if isinstance(arg, PPDLDeltaTerm)
    )


class PPDLDeltaSymbol(Symbol):
    def __init__(self, dist_name, n_terms):
        self.dist_name = dist_name
        self.n_terms = n_terms
        super().__init__(f"Result_{self.dist_name}_{self.n_terms}")

    def __repr__(self):
        return (
            f"Δ-Symbol{{{self.name}({self.dist_name}, "
            "{self.n_terms}): {self.type}}}"
        )

    def __hash__(self):
        return hash((self.dist_name, self.n_terms))


class PPDLDeltaTerm(FunctionApplication):
    def __repr__(self):
        return f"Δ-term{{{self.functor}({self.args}): {self.type}}}"


class PPDLMixin(PatternWalker):
    @add_match(
        Implication(FunctionApplication, ...),
        lambda exp: any(
            isinstance(arg, PPDLDeltaTerm) for arg in exp.consequent.args
        ),
    )
    def ppdl_rule(self, rule):
        if not is_ppdl_rule(rule):
            raise ForbiddenExpressionError(f"Invalid PPDL rule: {rule}")
        pred_symb = rule.consequent.functor
        pred_symb = pred_symb.cast(UnionOfConjunctiveQueries)
        if pred_symb in self.protected_keywords:
            raise ForbiddenExpressionError(f"symbol {pred_symb} is protected")
        if pred_symb in self.symbol_table:
            formulas = self.symbol_table[pred_symb].formulas
        else:
            formulas = tuple()
        formulas += (rule,)
        self.symbol_table[pred_symb] = Union(formulas)
        return rule


class PPDL(PPDLMixin, DatalogProgram, ExpressionWalker):
    """
    Class implementing a Probabilistic Programming DataLog (PPDL) program
    """

    pass



from typing import AbstractSet

from ...expression_pattern_matching import add_match
from ...expression_walker import PatternWalker
from ...expressions import (
    Constant,
    Definition,
    ExpressionBlock,
    FunctionApplication,
    Symbol,
)
from ...logic import Implication
from ...logic.expression_processing import extract_logic_predicates
from ...relational_algebra import (
    ConcatenateConstantColumn,
    Projection,
    RelationalAlgebraSolver,
    str2columnstr_constant,
)
from ..expressions import (
    Grounding,
    ProbabilisticChoiceGrounding,
    ProbabilisticPredicate,
)
from .grounding import get_grounding_pred_symb, topological_sort_groundings


class GraphicalModel(Definition):
    def __init__(self, nodes, edges):
        self.nodes = nodes
        self.edges = edges

    def get_node(self, node_symbol):
        nodes_as_dict = dict(self.nodes)
        if node_symbol not in nodes_as_dict:
            raise KeyError(f"Node {node_symbol} not found")
        return nodes_as_dict[node_symbol]

    def get_parent_node_symbols(self, node_symbol):
        return set(
            edge_parent_node_symb
            for edge_node_symb, edge_parent_node_symb in self.edges
            if edge_node_symb == node_symbol
        )


class PlateNode(Definition):
    def __init__(self, node_symbol, expression, relation):
        self.node_symbol = node_symbol
        self.expression = expression
        self.relation = relation

    def __repr__(self):
        return f"{self.node_symbol} ~ {self.__class__.__name__}"


class ProbabilisticPlateNode(PlateNode):
    """
    Object used to represent non-deterministic nodes.represent
    non-deterministic nodes

    Although this object is called a "CPD" factory, there is no parent
    to this distribution. Instead, it represents the marginal
    distributions of a batch of independent Bernoulli random variables.

    An extra column in the relation used to represent the random
    variables is used to encode the probability of that specific
    random variable's Bernoulli distribution:

    For example, to represent the program

        P(a) : 0.7 <- T
        P(b) : 0.2 <- T

    One would use the following relation

         _p_ | x
        =====|===
         0.7 | a
         0.2 | b

    where each row represents one random variable P(x) and where
    the column _p_ encodes the Bernoulli parameter of the
    distribution of P(x).

    Attributes
    ----------
    relation: Constant[AbstractSet]
        A relation that encodes all the variables for which a
        conditional probability distribution will be produced.
    probability_column: Constant[ColumnStr]
        The column that contains the Bernoulli parameter associated
        to each random variable.

    """

    def __init__(self, node_symbol, expression, relation, probability_column):
        super().__init__(node_symbol, expression, relation)
        self.probability_column = probability_column


class BernoulliPlateNode(ProbabilisticPlateNode):
    """
    Object used to produce Bernoulli-distributed random variables.

    Although this object is called a "CPD" factory, there is no parent
    to this distribution. Instead, it represents the marginal
    distributions of a batch of independent Bernoulli random variables.

    An extra column in the relation used to represent the random
    variables is used to encode the probability of that specific
    random variable's Bernoulli distribution:

    For example, to represent the program

        P(a) : 0.7 <- T
        P(b) : 0.2 <- T

    One would use the following relation

         _p_ | x
        =====|===
         0.7 | a
         0.2 | b

    where each row represents one random variable P(x) and where
    the column _p_ encodes the Bernoulli parameter of the
    distribution of P(x).

    Attributes
    ----------
    relation: Constant[AbstractSet]
        A relation that encodes all the variables for which a
        conditional probability distribution will be produced.
    probability_column: Constant[ColumnStr]
        The column that contains the Bernoulli parameter associated
        to each random variable.

    """


class AndPlateNode(PlateNode):
    """
    Object used to produce deterministic conditional probabilities
    that conjunct the values of boolean parent random variables.

    It only makes sense to use this CPD if all the parents are
    boolean random variables and if the random variable itself
    is a boolean random variable.

    Let X be the random variable for which the CPD is calculated
    and let Y_1, ..., Y_n be n parent boolean random variables.

    The AND CPD encodes that
                                              |- 1  if all y_i = T
        P(X = T | Y_1 = y_1, ... Y_n = y_1) = |
                                              |- 0  otherwise

    """


class NaryChoicePlateNode(ProbabilisticPlateNode):
    """
    Object used to represent the distribution of a single
    n-ary choice random variable.

    The relation does not represent multiple random variables
    but the different predicates that can be chosen by the rule,
    alongside their associated probability.

    Given a CP-Event of the form

        P(a_1) : p_1  v  ...  v  P(a_n) : p_n  <-  T

    where p_1, ..., p_n are the probability labels attached to each
    head predicate P(a_1), ..., P(a_n) -- respectively,

    the conditional probability distribution of the choice random
    variable (which is its marginal because there is no parent) is
    defined by the relation

        _p_ | x
        ====|====
        p_1 | a_1
        p_2 | a_2
        ... | ...
        p_n | a_n

    where each row represents a possible value of the random variable
    with its probability.

    """


class NaryChoiceResultPlateNode(PlateNode):
    pass


def is_extensional_grounding(grounding):
    """TODO: represent extensional grounding with a fact instead?"""
    return (
        isinstance(grounding.expression, Implication)
        and isinstance(grounding.expression.consequent, FunctionApplication)
        and grounding.expression.antecedent == Constant[bool](True)
    )


class CPLogicGroundingToGraphicalModelTranslator(PatternWalker):
    def __init__(self):
        self.edges = []
        self.nodes = []

    @add_match(ExpressionBlock)
    def block_of_groundings(self, block):
        if any(not isinstance(exp, Grounding) for exp in block.expressions):
            raise ValueError("Expected block of groundings")
        for grounding in topological_sort_groundings(block.expressions):
            self.walk(grounding)
        return GraphicalModel(tuple(self.nodes), tuple(self.edges))

    @add_match(Grounding, is_extensional_grounding)
    def extensional_set_grounding(self, grounding):
        """
        Represent a set of ground facts as a Bernoulli node with
        all probabilities set to 1.0.

        """
        expression = grounding.expression
        node_symbol = get_grounding_pred_symb(expression)
        probability_column = str2columnstr_constant(Symbol.fresh().name)
        relation = ConcatenateConstantColumn(
            grounding.relation, probability_column, Constant[float](1.0)
        )
        relation = RelationalAlgebraSolver().walk(relation)
        node = BernoulliPlateNode(
            node_symbol, expression, relation, probability_column
        )
        self.add_plate_node(node_symbol, node)

    @add_match(ProbabilisticChoiceGrounding)
    def probabilistic_choice_grounding(self, grounding):
        """
        Represent a probabilistic choice as a n-ary choice node.
        """
        expression = grounding.expression
        choice_node_symb = Symbol.fresh()
        probability_column = str2columnstr_constant(
            grounding.expression.consequent.probability.name
        )
        relation = grounding.relation
        # add a n-ary choice random variable
        choice_node = NaryChoicePlateNode(
            choice_node_symb, expression, relation, probability_column
        )
        self.add_plate_node(choice_node_symb, choice_node)
        # remove the probability column as it is not neeeded to represent the
        # CPD factories of boolean random variables whose value is
        # deterministically determined by the value of their parent choice
        # variable
        relation = RelationalAlgebraSolver().walk(
            Projection(
                relation,
                tuple(
                    str2columnstr_constant(col)
                    for col in relation.value.columns
                    if col != probability_column.value
                ),
            )
        )
        node_symbol = get_grounding_pred_symb(expression)
        node = NaryChoiceResultPlateNode(node_symbol, expression, relation)
        self.add_plate_node(
            node_symbol, node, parent_node_symbols={choice_node_symb}
        )

    @add_match(Grounding(Implication(ProbabilisticPredicate, ...), ...))
    def probfact_set_grounding(self, grounding):
        """
        Represent a set of probabilistic facts with a Bernoulli node.
        """
        expression = grounding.expression
        node_symbol = get_grounding_pred_symb(expression)
        probability_column = str2columnstr_constant(
            grounding.expression.consequent.probability.name
        )
        relation = grounding.relation
        node = BernoulliPlateNode(
            node_symbol, expression, relation, probability_column
        )
        self.add_plate_node(node_symbol, node)

    @add_match(Grounding(Implication, Constant[AbstractSet]))
    def intensional_rule_grounding(self, grounding):
        """
        Represent a deterministic intensional rule with an AND node.
        """
        expression = grounding.expression
        node_symbol = get_grounding_pred_symb(expression)
        parent_node_symbols = set(
            predicate.functor
            for predicate in extract_logic_predicates(
                grounding.expression.antecedent
            )
        )
        node = AndPlateNode(node_symbol, expression, grounding.relation)
        self.add_plate_node(node_symbol, node, parent_node_symbols)

    def add_plate_node(self, node_symbol, node, parent_node_symbols=None):
        self.nodes.append((node_symbol, node))
        if parent_node_symbols is not None:
            self.edges += [
                (node_symbol, parent_node_symb)
                for parent_node_symb in parent_node_symbols
            ]



from neurolang.probabilistic.expression_processing import is_probabilistic_fact
from ..expressions import Constant, Symbol
from ..relational_algebra import RelationalAlgebraOperation, ColumnInt


class ProbabilisticFactSet(RelationalAlgebraOperation):
    def __init__(self, relation, probability_column):
        self.relation = relation
        self.probability_column = probability_column


class ProbabilisticChoiceSet(RelationalAlgebraOperation):
    def __init__(self, relation, probability_column):
        self.relation = relation
        self.probability_column = probability_column


class DeterministicFactSet(RelationalAlgebraOperation):
    def __init__(self, relation):
        self.relation = relation


def generate_probabilistic_symbol_table_for_query(
    cpl_program, query_predicate
):
    """Generate a symbol table adding a fresh symbol for
    each probabilistic/deterministic table and wrapping
    it into one of the classes
    DeterministicFactSet, ProbabilisticFactSet, ProbabilisticChoiceSet

    Parameters
    ----------
    cpl_program : neurolang.probabilistic.cplogic.CPLogicProgram
        CPLogic program representation indexing the query's symbols.
    query_predicate : neurolang.logic.LogicOperator
        Logic expression representing a query from where
        to extract the predicates which are represented as Symbols.

    Returns
    -------
    dict or Mapping
        symbol table with relationship symbols of query_predicate.
        The tables has neurolang.expressions.Symbol as keys and
        relational algebra representations as values
    """
    symbol_table = dict()
    classify_and_wrap_symbols(
        cpl_program.probabilistic_facts(), query_predicate,
        symbol_table, ProbabilisticFactSet
    )

    classify_and_wrap_symbols(
        cpl_program.probabilistic_choices(), query_predicate,
        symbol_table, ProbabilisticChoiceSet
    )

    classify_and_wrap_symbols(
        cpl_program.extensional_database(), query_predicate,
        symbol_table, lambda s, _: DeterministicFactSet(s)
    )

    return symbol_table


def classify_and_wrap_symbols(
    ra_set_dict, query_predicate, symbol_table, wrapper
):
    for predicate_symbol, facts in ra_set_dict.items():
        if predicate_symbol not in query_predicate._symbols:
            continue

        fresh_symbol = Symbol.fresh()
        symbol_table[predicate_symbol] = wrapper(
            fresh_symbol,
            Constant[ColumnInt](ColumnInt(0))
        )
        symbol_table[fresh_symbol] = facts



import itertools
import typing
from collections import OrderedDict
from operator import add, and_, invert, mul, neg, or_, pos, sub, truediv
from warnings import warn

from .exceptions import NeuroLangException
from .expression_walker import (ExpressionBasicEvaluator, PatternWalker,
                                ReplaceSymbolWalker, TypedSymbolTableEvaluator,
                                add_match)
from .expressions import (Constant, Definition, Expression,
                          FunctionApplication, NonConstant, Query, Symbol,
                          Unknown, is_leq_informative)
from .logic import ExistentialPredicate, UniversalPredicate

T = typing.TypeVar('T')


class NeuroLangPredicateException(NeuroLangException):
    pass


class GenericSolver(ExpressionBasicEvaluator, TypedSymbolTableEvaluator):
    @property
    def plural_type_name(self):
        return self.type_name + 's'

    def set_symbol_table(self, symbol_table):
        self.symbol_table = symbol_table


class BooleanRewriteSolver(PatternWalker):
    @add_match(
       FunctionApplication(Constant, (Expression[bool],) * 2),
       lambda expression: (
           expression.functor.value in (or_, and_) and
           expression.type is not bool
        )
    )
    def cast_binary(self, expression):
        functor, args = expression.functor, expression.args
        new_functor = functor.cast(typing.Callable[[bool, bool], bool])
        new_application = FunctionApplication[bool](new_functor, args)
        return self.walk(new_application)

    @add_match(
       FunctionApplication(Constant(invert), (Expression[bool],)),
       lambda expression: (
           expression.type is not bool
        )
    )
    def cast_unary(self, expression):
        functor, args = expression.functor, expression.args
        new_functor = functor.cast(typing.Callable[[bool], bool])
        new_application = FunctionApplication[bool](new_functor, args)
        return self.walk(new_application)

    @add_match(
        FunctionApplication[bool](
            Constant(invert), (
                FunctionApplication[bool](
                    Constant(invert), (
                        Expression[bool],
                    )
                ),
            )
        )
    )
    def simplify_double_inversion(self, expression):
        return self.walk(expression.args[0].args[0])

    @add_match(
        FunctionApplication[bool](
            Constant(...),
            (NonConstant[bool], Constant[bool])
        ),
        lambda expression: (
            expression.functor.value in (or_, and_)
        )
    )
    def dual_operator(self, expression):
        return self.walk(
            FunctionApplication[bool](
                expression.functor,
                expression.args[::-1]
            )
        )

    @add_match(
        FunctionApplication[bool](Constant, (
            NonConstant[bool],
            FunctionApplication[bool](
                Constant,
                (Constant[bool], Expression[bool])
            )
        )),
        lambda expression: (
            expression.functor.value in (or_, and_)
            and expression.args[1].functor.value is expression.functor.value
        )
    )
    def bring_constants_up_left(self, expression):
        outer = expression
        inner = expression.args[1]
        new_inner = FunctionApplication[bool](
            inner.functor, (outer.args[0], inner.args[1])
        )
        new_outer = FunctionApplication[bool](
            outer.functor, (inner.args[0], new_inner)
        )
        return self.walk(new_outer)

    @add_match(
        FunctionApplication[bool](
            Constant(invert),
            (FunctionApplication[bool](
                Constant(or_),
                (Expression[bool], Expression[bool])
            ),)
        )
    )
    def neg_disj_to_conj(self, expression):
        return self.walk(
            FunctionApplication[bool](
                Constant(and_), (
                    (~expression.args[0].args[0]).cast(bool),
                    (~expression.args[0].args[1]).cast(bool)
                )
            )
        )

    @add_match(
        FunctionApplication[bool](
            Constant(...), (FunctionApplication[bool], Expression[bool])
        ),
        lambda expression: expression.functor.value in (or_, and_) and
        any(
            isinstance(arg, Definition) for arg in expression.args[0].args
        ) and (
            not isinstance(expression.args[1], Definition) or (
                all(
                    not isinstance(arg, Definition)
                    for arg in expression.args[1].args
                )
            )
        )
    )
    def conjunction_composition_dual(self, expression):
        return self.walk(
            FunctionApplication[bool](
                Constant(expression.functor.value),
                (expression.args[1], expression.args[0])
            )
        )

    @add_match(
        FunctionApplication[bool](
            Constant(...), (Definition, Expression[bool])
        ),
        lambda expression: expression.functor.value in (or_, and_) and
        not isinstance(expression.args[1], Definition)
    )
    def conjunction_definition_dual(self, expression):
        return self.walk(
            FunctionApplication[bool](
                Constant(expression.functor.value),
                (expression.args[1], expression.args[0])
            )
        )

    @add_match(
        FunctionApplication(
            Constant(and_),
            (
                FunctionApplication(Constant(and_), ...),
                ...
            )
        )
    )
    def conjunction_distribution(self, expression):
        return self.walk(FunctionApplication[expression.type](
            expression.functor,
            (
                expression.args[0].args[0],
                FunctionApplication[expression.args[0].type](
                    expression.args[0].functor,
                    (expression.args[0].args[1], expression.args[1])
                )
            )
        ))

    @add_match(
        FunctionApplication(Constant(...), (NonConstant, NonConstant)),
        lambda expression: expression.functor.value in (or_, and_)
    )
    def partial_binary_evaluation(self, expression):
        '''Partially evaluate a binary AND or OR operator with
        non-constant arguments by giving priority to the evaluation of
        the first argument.

        Parameters
        ----------
        expression : FunctionApplication
            A OR or AND binary operator application.

        Returns
        -------
        FunctionApplication
            Same expression passed as parameter where the first
            argument was walked on and the second argument was
            potentially walked on if the walk on the first argument
            did not modify its expression.

        '''
        first_arg = expression.args[0]
        # we walk on the first argument
        walk_first_result = self.walk(first_arg)
        # and replace that argument with the result of the walk
        new_args = (walk_first_result, expression.args[1])
        # if the walk on the first argument did not change anything
        # we walk on the second argument and replace it with the result
        if walk_first_result is first_arg:
            new_args = (first_arg, self.walk(expression.args[1]))

        # if the expression arguments did not change, we stop walking here
        if new_args == expression.args:
            return expression
        # otherwise, we keep walking
        else:
            return self.walk(
                FunctionApplication[bool](expression.functor, new_args)
            )


class BooleanOperationsSolver(PatternWalker):
    @add_match(FunctionApplication(Constant(invert), (Constant[bool],)))
    def rewrite_boolean_inversion(self, expression):
        return Constant(not expression.args[0].value)

    @add_match(
        FunctionApplication(Constant(and_), (Constant[bool], Constant[bool]))
    )
    def rewrite_boolean_and(self, expression):
        return Constant(expression.args[0].value and expression.args[1].value)

    @add_match(
        FunctionApplication(Constant(or_), (Constant[bool], Constant[bool]))
    )
    def rewrite_boolean_or(self, expression):
        return Constant(expression.args[0].value or expression.args[1].value)

    @add_match(
        FunctionApplication(Constant(or_), (True, Expression[bool]))
    )
    def rewrite_boolean_or_l(self, expression):
        return Constant(True)

    @add_match(
        FunctionApplication(Constant(or_), (Expression[bool], True))
    )
    def rewrite_boolean_or_r(self, expression):
        return Constant(True)

    @add_match(
        FunctionApplication(Constant(and_), (False, Expression[bool]))
    )
    def rewrite_boolean_and_l(self, expression):
        return Constant(False)

    @add_match(
        FunctionApplication(Constant(and_), (Expression[bool], False))
    )
    def rewrite_boolean_and_r(self, expression):
        return Constant(False)


class NumericOperationsSolver(PatternWalker[T]):
    @add_match(
        FunctionApplication[Unknown](Constant, (Expression[T],) * 2),
        lambda expression: expression.functor.value in (add, sub, mul, truediv)
    )
    def cast_binary(self, expression):
        type_ = expression.args[0].type
        functor = expression.functor.cast(
            typing.Callable[[type_, type_], type_]
        )
        if functor is not expression.functor:
            new_expression = FunctionApplication[type_](
                functor, expression.args
            )
        else:
            new_expression = expression.cast(type_)
        return self.walk(new_expression)

    @add_match(
        FunctionApplication(Constant, (Expression[T],)),
        lambda expression: expression.functor.value in (pos, neg)
    )
    def cast_unary(self, expression):
        return self.walk(expression.cast(expression.args[0].type))


class FirstOrderLogicSolver(
        BooleanRewriteSolver,
        BooleanOperationsSolver,
        NumericOperationsSolver[int],
        NumericOperationsSolver[float],
        GenericSolver
):
    '''
    WIP non-recursive first order logic query solver.
    For now predicates work only on constants on the symbols table
    '''

    @add_match(
        Query,
        guard=lambda expression: (
            expression.head._symbols == expression.body._symbols
        )
    )
    def query_resolution(self, expression):
        out_query_type = expression.type
        if out_query_type is Unknown:
            out_query_type = typing.AbstractSet[expression.head.type]

        result = []

        symbols_domains = self.quantifier_head_symbols_and_adom(
            expression.head
        )
        for symbol_values in itertools.product(*symbols_domains.values()):
            rsw = ReplaceSymbolWalker(
                dict(zip(
                    symbols_domains.keys(),
                    (s[1] for s in symbol_values)
                ))
            )
            body = rsw.walk(expression.body)

            res = self.walk(body)
            if isinstance(res, Constant):
                if res.value is True:
                    if isinstance(expression.head, Symbol):
                        result.append(symbol_values[0][0])
                    else:
                        result.append(tuple(zip(*symbol_values))[0])
            else:
                warn('Query body could not be evaluated')

        return Constant[out_query_type](
            frozenset(result)
        )

    @add_match(
        ExistentialPredicate,
        lambda expression: expression.head._symbols == expression.body._symbols
    )
    def existential_predicate(self, expression):
        symbols_domains = self.quantifier_head_symbols_and_adom(
            expression.head
        )

        for symbol_values in itertools.product(*symbols_domains.values()):
            rsw = ReplaceSymbolWalker(
                dict(zip(
                    symbols_domains.keys(),
                    (s[1] for s in symbol_values)
                ))
            )
            body = rsw.walk(expression.body)
            res = self.walk(body)
            if isinstance(res, Constant) and res.value:
                return Constant(True)

        return Constant(False)

    @add_match(
        UniversalPredicate,
        lambda expression: expression.head._symbols == expression.body._symbols
    )
    def universal_predicate(self, expression):
        symbols_domains = self.quantifier_head_symbols_and_adom(
            expression.head
        )
        for symbol_values in itertools.product(*symbols_domains.values()):
            rsw = ReplaceSymbolWalker(
                dict(zip(
                    symbols_domains.keys(),
                    (s[1] for s in symbol_values)
                ))
            )
            body = rsw.walk(expression.body)
            res = self.walk(body)
            if not res.value:
                return Constant(False)

        return Constant(True)

    def quantifier_head_symbols_and_adom(self, head):
        '''
        Returns an ordered dictionary with the symbols of the quantifier head
        as keys and the active domain for each symbol as value.
        '''
        if (
            isinstance(head, Constant) and
            is_leq_informative(head.type, typing.Tuple) and
            all(isinstance(a, Symbol) for a in head.value)
        ):
            symbols_in_head = head.value
        else:
            symbols_in_head = (head,)

        constants = tuple((
            (
                (k, v)
                for k, v in self.symbol_table.symbols_by_type(
                    sym.type
                ).items()
                if isinstance(v, Constant)
            )
            for sym in symbols_in_head
        ))
        return OrderedDict(zip(symbols_in_head, constants))

    @staticmethod
    def new_set(iterable):
        return frozenset(iterable)



class NeuroLangException(Exception):
    """Base class for NeuroLang Exceptions"""

    pass


class UnexpectedExpressionError(NeuroLangException):
    pass


class NeuroLangNotImplementedError(NeuroLangException):
    pass


class ForbiddenExpressionError(NeuroLangException):
    pass


class ForbiddenDisjunctionError(ForbiddenExpressionError):
    pass


class ForbiddenExistentialError(ForbiddenExpressionError):
    pass


class RelationalAlgebraError(NeuroLangException):
    pass


class RelationalAlgebraNotImplementedError(
    RelationalAlgebraError, NotImplementedError
):
    pass


class ForbiddenBuiltinError(ForbiddenExpressionError):
    pass


class NeuroLangFrontendException(NeuroLangException):
    pass


class SymbolNotFoundError(NeuroLangException):
    pass


class RuleNotFoundError(NeuroLangException):
    pass


class UnsupportedProgramError(NeuroLangException):
    pass


class UnsupportedQueryError(NeuroLangException):
    pass


class ProtectedKeywordError(NeuroLangException):
    pass


class ForbiddenRecursivityError(UnsupportedProgramError):
    pass


class ForbiddenUnstratifiedAggregation(UnsupportedProgramError):
    pass



import numpy as np

from ..perior_tree import BoundedAABB, Boundary, Tree

def _generate_random_box(x_bounds, y_bounds, boundry, size_bounds):
    lower_bound = np.array([np.random.uniform(*b) for b in (x_bounds, y_bounds)])
    upper_bound = lower_bound + np.random.uniform(*size_bounds, size=2)
    return BoundedAABB(lower_bound, upper_bound, boundry)

def test_point_adjust_position():
    period_bound = Boundary((0, 0), (10, 10))
    point = np.asanyarray((-3, -7))
    inbound_point = period_bound.adjust_position(point)
    assert inbound_point == (7, 3)

def test_vector_adjust_direction():
    period_bound = Boundary((0, 0), (10, 10))
    point = np.asanyarray((9.5, 0.5))
    dir_vec = period_bound.adjust_direction(point)
    assert dir_vec == (-0.5, 0.5)


    # #TODO: refa ploting
    # rect = plt.Rectangle(period_bound[0], period_bound.width(), period_bound.height(), facecolor="#aaaaaa")
    # fig, ax = plt.subplots()
    # ax.add_patch(rect)
    #
    # plt.plot(point[0], point[1], 'ro')
    # plt.show()


def test_adjust_aabb_out_of_bound():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((-3, -3), (-1,-1), period_bound)
    box1.adjust_to_bound()
    assert box1 == BoundedAABB((7, 7), (9, 9), period_bound)

    # fig, ax = plt.subplots()
    # ax.add_patch(patches.Rectangle(box1[0], box1.width(), box1.height(), hatch='+', fill=False))
    # ax.axis([period_bound[0][0], period_bound[1][0], period_bound[0][1], period_bound[1][1]])
    # plt.show()

def test_aabbs_union_in_bound():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((1, 1), (3, 3), period_bound)
    box2 = BoundedAABB((5, 6), (6, 8), period_bound)
    assert box1.expand(box2) == BoundedAABB((1, 1), (6, 8), period_bound)


    # #Todo: refa ploting
    # fig, ax = plt.subplots()
    # ax.add_patch(patches.Rectangle(box1[0], box1.width(), box1.height(), hatch='+', fill=False))
    # ax.add_patch(patches.Rectangle(box2[0], box2.width(), box2.height(), hatch='*', fill=False))
    # ax.axis([0, 10, 0, 10])
    # plt.show()

def test_contains_aabbs():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((1, 1), (3, 3), period_bound)
    box2 = BoundedAABB((0, 0), (2,2), period_bound)
    assert not box1.contains(box2)

    box1 = BoundedAABB((1, 1), (3, 3), period_bound)
    box2 = BoundedAABB((1.5, 1.5), (2,2), period_bound)
    assert box1.contains(box2)

    box1 = BoundedAABB((1, 1), (3, 3), period_bound)
    box2 = BoundedAABB((1, 1), (3, 3), period_bound)
    assert box1.contains(box2)

    box1 = BoundedAABB((-10, -10), (0, 0), period_bound)
    box2 = BoundedAABB((1, 1), (3, 3), period_bound)
    assert box1.contains(box2)


def test_aabbs_intersect():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((1, 1), (3, 3), period_bound)
    box2 = BoundedAABB((2, 2), (4, 4), period_bound)
    assert box1.intersects(box1)
    assert box1.intersects(box2)
    assert box2.intersects(box1)

    box1 = BoundedAABB((1, 1), (2, 2), period_bound)
    box2 = BoundedAABB((3, 3), (4, 4), period_bound)
    assert not box1.intersects(box2)

    box1 = BoundedAABB((-2, -2), (0, 0), period_bound)
    box2 = BoundedAABB((9, 9), (10, 10), period_bound)
    assert box1.intersects(box2)


def test_expand_aabb_point():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((0, 0), (5, 5), period_bound)
    assert box1.expand((7, 7)) == BoundedAABB((0, 0), (7, 7), period_bound)

def test_expand_aabb_point_outbound():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((9, 9), (10, 10), period_bound)
    assert box1.expand((14, 9)) == BoundedAABB((9, 9), (4, 9), period_bound)

def test_expand_aabbs_inbound():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((0, 0), (5, 5), period_bound)
    box2 = BoundedAABB((2, 2), (7, 7), period_bound)
    assert box1.expand(box2) == BoundedAABB((0, 0), (7, 7), period_bound)


def test_expand_aabbs_outside_bound():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((0, 0), (1, 1), period_bound)
    box2 = BoundedAABB((9, 0), (10, 1), period_bound)
    assert box1.expand(box2) == BoundedAABB((9, 0), (1, 1), period_bound)


def test_boundry_eq():
    period_bound = Boundary((0, 0), (10, 10))
    box1 = BoundedAABB((0, 0), (5, 5), period_bound)
    box2 = BoundedAABB((0, 0), (5, 5), period_bound)
    assert box1 == box2


def test_tree_construction():
    tree = Tree()
    assert tree.root is None


def test_tree_add():
    period_bound = Boundary((0, 0), (10, 10))
    tree = Tree()
    box1 = BoundedAABB((0, 0), (1, 1), period_bound)
    box2 = BoundedAABB((0.5, 0.5), (2, 2), period_bound)
    tree.add(box1)
    assert tree.root is not None
    tree.add(box2)
    assert tree.root.box == box1.expand(box2)
    assert tree.root.left.box == box1
    assert tree.root.right.box == box2

    for _ in range(100):
        tree.add(_generate_random_box((-2, -1), (0, 1), period_bound, (0.2, 0.7)))
        tree.add(_generate_random_box((1, 2), (0, 1), period_bound, (0.2, 0.7)))

def test_tree_query_regions_contained_in_box():
    period_bound = Boundary((0, 0), (10, 10))
    tree = Tree()
    tree.add(BoundedAABB((2, 2), (3, 3), period_bound), region_ids={1})
    box = BoundedAABB((1, 1), (4, 4), period_bound)
    assert tree.query_regions_contained_in_box(box) == {1}
    box = BoundedAABB((2.5, 2.5), (4, 4), period_bound)
    assert tree.query_regions_contained_in_box(box) == set()



import operator as op
from typing import Callable

from .. import expressions
from ..expression_walker import (ExpressionBasicEvaluator, ReplaceSymbolWalker,
                                 TypedSymbolTableEvaluator)

C_ = expressions.Constant
S_ = expressions.Symbol


class Evaluator(
    TypedSymbolTableEvaluator,
    ExpressionBasicEvaluator
):
    pass


def test_replace_in_walker():
    value = C_[int](2)
    symbol_to_replace = S_('x')

    rsw = ReplaceSymbolWalker({symbol_to_replace: value})
    result = rsw.walk(symbol_to_replace)

    assert result == value


def test_replace_variable_in_expression():
    symbol_to_replace = S_('a')
    value = C_[int](2)

    add_constant = C_[Callable[[int, int], int]](op.add)
    add_op = add_constant(symbol_to_replace, C_[int](3))

    rsw = ReplaceSymbolWalker({symbol_to_replace: value})
    add_replacement = rsw.walk(add_op)

    ebe = Evaluator()
    add_result = ebe.walk(add_replacement)
    assert add_result == 5


def test_replace_variables_in_expression():
    symbols_to_replace = {
        S_('a'): C_[int](2),
        S_('b'): C_[int](3)
    }

    add_constant = C_[Callable[[int, int], int]](op.add)
    add_op = add_constant(*symbols_to_replace)

    rsw = ReplaceSymbolWalker(symbols_to_replace)
    add_replacement = rsw.walk(add_op)

    ebe = Evaluator()
    add_result = ebe.walk(add_replacement)
    assert add_result == 5



from operator import eq

import pytest

from ... import exceptions
from ... import expression_walker as ew
from ... import expressions
from ... import logic
from .. import Fact, Implication
from .. import negation as sdn
from ..expressions import TranslateToLogic
from ..chase import Chase as Chase_
from ..chase.negation import NegativeFactConstraints, DatalogChaseNegation


C_ = expressions.Constant
S_ = expressions.Symbol
F_ = expressions.FunctionApplication
L_ = expressions.Lambda
E_ = logic.ExistentialPredicate
U_ = logic.UniversalPredicate
Eb_ = expressions.ExpressionBlock


class Datalog(
    TranslateToLogic,
    sdn.DatalogProgramNegation,
    ew.ExpressionBasicEvaluator
):
    def function_gt(self, x: int, y: int) -> bool:
        return x > y


class Chase(NegativeFactConstraints, Chase_):
    pass


def test_non_recursive_negation():
    x = S_('x')
    G = S_('G')
    T = S_('T')
    V = S_('V')

    program = Eb_((
        Fact(V(C_(1))),
        Fact(V(C_(2))),
        Fact(V(C_(3))),
        Fact(T(C_(1))),
        Fact(T(C_(4))),
        Implication(G(x), V(x) & ~T(x))
    ))

    dl = Datalog()
    dl.walk(program)

    dc = Chase(dl)
    solution_instance = dc.build_chase_solution()

    assert solution_instance['V'].value == {(1,), (2,), (3,)}
    assert solution_instance['T'].value == {(1,), (4,)}
    assert solution_instance['G'].value == {(2,), (3,)}


def test_stratified_and_chase():
    x = S_('x')
    y = S_('y')
    z = S_('z')
    G = S_('G')
    T = S_('T')
    V = S_('V')
    ET = S_('ET')
    equals = S_('equals')

    program = Eb_((
        Fact(V(C_(1))),
        Fact(V(C_(2))),
        Fact(V(C_(3))),
        Fact(V(C_(4))),
        Fact(G(C_(1), C_(2))),
        Fact(G(C_(2), C_(3))),
        Implication(T(x, y),
                    V(x) & V(y) & G(x, y)),
        Implication(T(x, y),
                    G(x, z) & T(z, y)),
        Implication(ET(x, y),
                    V(x) & V(y) & ~(G(x, y)) & equals(x, y)),
    ))

    dl = Datalog()
    dl.walk(program)

    #dc = DatalogChaseNegation(dl)
    dc = Chase(dl)
    solution_instance = dc.build_chase_solution()

    final_instance = {
        G:
        C_({
            C_((C_(1), C_(2))),
            C_((C_(2), C_(3))),
        }),
        V:
        C_({
            C_((C_(2), )),
            C_((C_(3), )),
            C_((C_(1), )),
            C_((C_(4), )),
        }),
        ET:
        C_({
            C_((C_(1), C_(1))),
            C_((C_(3), C_(3))),
            C_((C_(4), C_(4))),
            C_((C_(2), C_(2))),
        }),
        T:
        C_({C_((C_(1), C_(2))),
            C_((C_(1), C_(3))),
            C_((C_(2), C_(3)))})
    }

    assert solution_instance == final_instance


def test_stratified_and_chase_builtin_equality():
    x = S_('x')
    y = S_('y')
    G = S_('G')
    V = S_('V')
    NT = S_('NT')
    equals = C_(eq)

    program = Eb_((
        Fact(V(C_(1))),
        Fact(V(C_(2))),
        Fact(V(C_(3))),
        Fact(V(C_(4))),
        Fact(G(C_(1), C_(2))),
        Fact(G(C_(2), C_(3))),
        Implication(
            NT(x, y),
            V(x) & V(y) & ~(G(x, y)) & ~(equals(x, y))
        ),
    ))

    dl = Datalog()
    dl.walk(program)

    dc = Chase(dl)
    solution_instance = dc.build_chase_solution()

    final_instance = {
        G:
        C_({
            C_((C_(1), C_(2))),
            C_((C_(2), C_(3))),
        }),
        V:
        C_({
            C_((C_(2), )),
            C_((C_(3), )),
            C_((C_(1), )),
            C_((C_(4), )),
        }),
        NT:
        C_({
            C_((C_(3), C_(2))),
            C_((C_(1), C_(3))),
            C_((C_(4), C_(1))),
            C_((C_(3), C_(1))),
            C_((C_(2), C_(1))),
            C_((C_(1), C_(4))),
            C_((C_(4), C_(3))),
            C_((C_(4), C_(2))),
            C_((C_(3), C_(4))),
            C_((C_(2), C_(4))),
        }),
    }

    assert solution_instance == final_instance


def test_negative_fact():
    V = S_('V')
    G = S_('G')
    T = S_('T')
    CT = S_('CT')
    x = S_('x')
    y = S_('y')
    z = S_('z')

    program = Eb_((
        Fact(V(C_(1))),
        Fact(V(C_(2))),
        Fact(V(C_(3))),
        Fact(V(C_(4))),
        Fact(G(C_(1), C_(2))),
        Fact(G(C_(2), C_(3))),
        sdn.NegativeFact(G(C_(1), C_(2))),
        Implication(T(x, y),
                    V(x) & V(y) & G(x, y)),
        Implication(T(x, y),
                    G(x, z) & T(z, y)),
        Implication(CT(x, y),
                    V(x) & V(y) & ~(G(x, y))),
    ))

    dl = Datalog()
    dl.walk(program)

    dc = Chase(dl)
    with pytest.raises(
        exceptions.NeuroLangException, match=r'There is a contradiction .*'
    ):
        dc.build_chase_solution()


def test_symbol_order_in_datalog_solver():
    n = S_("n")
    l = S_("l")
    R = S_("R")
    S = S_("S")
    Ans = S_("Ans")

    program = Eb_((Implication(Ans(l), R(l, n) & ~S(n, l)),))

    dl = Datalog()
    dl.walk(program)
    dl.add_extensional_predicate_from_tuples(
        R, {("a", 1), ("a", 2), ("a", 3), ("b", 3), ("b", 4)}
    )
    dl.add_extensional_predicate_from_tuples(
        S, {(1, "a"), (2, "a"), (3, "b"), (4, "b")}
    )

    dc = Chase(dl)
    solution_instance = dc.build_chase_solution()

    assert solution_instance["Ans"].value == {("a",)}


def test_negative_predicates_race_in_chase():
    x = S_("x")
    F = S_("F")
    G = S_("G")
    R = S_("R")
    S = S_("S")

    dl = Datalog()
    dl.add_extensional_predicate_from_tuples(
        F, {(1,), (2,), (3,)}
    )
    dl.add_extensional_predicate_from_tuples(
        G, {(1,), (2,), (3,), (4,), (5,)}
    )
    program = Eb_((
        Implication(R(x), F(x)),
        Implication(S(x), G(x) & ~R(x)),
    ))
    dl.walk(program)

    dc = DatalogChaseNegation(dl)
    solution_instance = dc.build_chase_solution()

    assert solution_instance["S"].value == {(4,), (5,)}



from ...expressions import Symbol
from ...logic import Conjunction
from ..dichotomy_theorem_based_solver import is_hierarchical_without_self_joins

Q = Symbol('Q')
R = Symbol('R')
S = Symbol('S')
T = Symbol('T')
x = Symbol('x')
y = Symbol('y')
z = Symbol('z')


def test_hierarchical_without_self_joins():
    q1 = Conjunction((Q(x), R(y), T(x)))
    q2 = Conjunction((Q(x),))
    q3 = Conjunction((Q(x), R(x, y)))
    q4 = Conjunction((Q(x), R(x, y, z), T(x, y)))

    assert is_hierarchical_without_self_joins(q1)
    assert is_hierarchical_without_self_joins(q2)
    assert is_hierarchical_without_self_joins(q3)
    assert is_hierarchical_without_self_joins(q4)


def test_not_hierarchical_without_self_joins():
    q1 = Conjunction((Q(x), T(x, y), R(y)))
    q2 = Conjunction((Q(x), Q(y), T(x, y)))

    assert not is_hierarchical_without_self_joins(q1)
    assert not is_hierarchical_without_self_joins(q2)



import collections

import numpy as np

from neurolang.logic.expression_processing import extract_logic_atoms

from ..datalog.expression_processing import (
    dependency_matrix,
    extract_logic_predicates,
    reachable_code,
)
from ..exceptions import ForbiddenRecursivityError, UnsupportedProgramError
from ..expressions import Symbol
from ..logic import Implication, Union


def _iter_implication_or_union_of_implications(expression):
    if expression is None:
        return
    if isinstance(expression, Implication):
        yield expression
    elif isinstance(expression, Union):
        for formula in expression.formulas:
            yield formula


def reachable_code_from_query(query, program):
    """
    Find the part of the intensional database of the program that is needed to
    answer a rule-based query Head :- Body such that Head is not necesarrily
    part of the program.

    This is useful if the query is formulated as a rule instead of simply a
    query predicate (e.g. `ans(x, y) :- P(x, z), P(y, z)`).

    """
    if query is None:
        return Union(tuple(_get_list_of_intensional_rules(program)))
    predicates = [query.consequent] + list(
        extract_logic_predicates(query.antecedent)
    )
    reachable = set()
    for pred in predicates:
        for rule in _iter_implication_or_union_of_implications(
            program.intensional_database().get(pred.functor, None)
        ):
            reachable |= set(reachable_code(rule, program).formulas)
    return Union(tuple(reachable))


def stratify_program(query, program):
    """
    Statically analyse the program to isolate its deterministic strat, its
    probabilistic strats and its post-probabilistic-query deterministic strat.

    A query can be solved through stratification if the probabilistic and
    deterministic parts are well separated. In case there exists one
    within-language probabilistic query dependency, no probabilistic predicate
    should appear in the strat that depends on the query.

    Parameters
    ----------
    query : Implication
        Query defining the part of the program that needs to be stratified.
    program : CPLogicProgram
        Program that will be stratified.

    Returns
    -------
    mapping from idb type to Union
        Deterministic, probabilistic and post-probabilistic IDBs.

    Raises
    ------
    UnsupportedProgramError
        When a WLQ (within-language query) depends on another WLQ.

    """
    idb = list(reachable_code_from_query(query, program).formulas)
    idb_symbs, dep_mat = dependency_matrix(program, idb)
    wlq_symbs = set(program.within_language_succ_queries()).intersection(
        idb_symbs
    )
    _check_for_dependencies_between_wlqs(dep_mat, idb_symbs, wlq_symbs)
    grpd_symbs = collections.defaultdict(set)
    grpd_symbs["deterministic"] = _get_program_deterministic_symbols(program)
    grpd_symbs["probabilistic"] = program.probabilistic_predicate_symbols
    grpd_idbs = collections.defaultdict(list)
    count = len(idb)
    while idb:
        rule = idb.pop(0)
        idb_type = _get_rule_idb_type(rule, grpd_symbs, wlq_symbs)
        if idb_type is None:
            idb.append(rule)
            count -= 1
            if count < 0:
                raise ForbiddenRecursivityError(
                    "Unstratifiable recursive program"
                )
        else:
            grpd_symbs[idb_type].add(rule.consequent.functor)
            grpd_idbs[idb_type].append(rule)
            count = len(idb)
    return {
        idb_type: Union(tuple(idb_rules))
        for idb_type, idb_rules in grpd_idbs.items()
    }


def _get_list_of_intensional_rules(program):
    idb = [
        rule
        for exp in program.intensional_database().values()
        for rule in _iter_implication_or_union_of_implications(exp)
    ]
    return idb


def _get_program_deterministic_symbols(program):
    det_symbs = set(program.extensional_database().keys())
    det_symbs |= set(program.builtins())
    if hasattr(program, "constraints"):
        det_symbs |= set(
            formula.consequent.functor
            for formula in program.constraints().formulas
        )
    return det_symbs


def _get_rule_idb_type(rule, grpd_symbs, wlq_symbs):
    dep_symbs = set(
        pred.functor
        for pred in extract_logic_atoms(rule.antecedent)
        if isinstance(pred.functor, Symbol)
    )
    idb_type = None
    # handle the case of a WLQ with deterministic-only dependencies
    if rule.consequent.functor in wlq_symbs:
        idb_type = "probabilistic"
    elif grpd_symbs["deterministic"].issuperset(dep_symbs):
        idb_type = "deterministic"
    elif (
        grpd_symbs["deterministic"]
        | wlq_symbs
        | grpd_symbs["post_probabilistic"]
    ).issuperset(dep_symbs):
        idb_type = "post_probabilistic"
    elif not grpd_symbs["probabilistic"].isdisjoint(dep_symbs):
        idb_type = "probabilistic"
    return idb_type


def _check_for_dependencies_between_wlqs(dep_mat, idb_symbs, wlq_symbs):
    """
    Raise an exception if a WLQ (Within-Language Query) depends on another WLQ.
    """
    wlq_symb_idxs = {idb_symbs.index(wlq_symb) for wlq_symb in wlq_symbs}
    for wlq_symb_idx in wlq_symb_idxs:
        if _wlq_depends_on_other_wlq(wlq_symb_idx, dep_mat, wlq_symb_idxs):
            raise UnsupportedProgramError(
                "Unsupported dependency between within-language queries"
            )


def _wlq_depends_on_other_wlq(wlq_symb_idx, dep_mat, wlq_symb_idxs):
    stack = [wlq_symb_idx]
    while stack:
        dep_idxs = np.argwhere(dep_mat[stack.pop()].astype(bool)).flatten()
        if not wlq_symb_idxs.isdisjoint(dep_idxs):
            return True
        stack += list(dep_idxs)
    return False



from numpy import random
from ..interval_algebra import (
    converse, meets, before, starts, during,
    finishes, equals, overlaps, negate,
    get_intervals_relations
)
from ..regions import Region
from copy import deepcopy


def app(x, z, first, second, elements):
    elements.remove(x)
    elements.remove(z)
    for y in elements:
        if first(x, y) and second(y, z):
            return True
    return False


def composition(relations, domain, convert=False):
    if convert:
        relations = [[converse(f), converse(g)] for [f, g] in relations]
    return [
        lambda x, z, f=pair_of_fs[0], g=pair_of_fs[1]:
        app(x, z, f, g, deepcopy(domain)) for pair_of_fs in relations
    ]


def apply_composition(
    relations,
    parameters,
    negate_whole_expression=False,
    negations=None,
    conversion=False
):

    if conversion:
        parameters = list(reversed(parameters))
    res = True
    for i in range(len(relations)):
        result = relations[i](parameters[i][0], parameters[i][1])
        if negations:
            if negations[i] != (not result):
                res = False
                break
        else:
            if not result:
                res = False
                break

    return negate_whole_expression != res


def test_ia_relations_functions():
    intervals = [
        tuple([1, 2]),
        tuple([5, 7]),
        tuple([1, 5]),
        tuple([4, 6]),
        tuple([2, 4]),
        tuple([6, 7]),
        tuple([2, 4])
    ]

    assert before(intervals[0], intervals[1])
    assert meets(intervals[0], intervals[4])
    assert starts(intervals[0], intervals[2])
    assert during(intervals[4], intervals[2])
    assert overlaps(intervals[3], intervals[1])
    assert finishes(intervals[5], intervals[1])
    assert equals(intervals[4], intervals[6])

    assert not equals(intervals[1], intervals[0])
    assert not during(intervals[1], intervals[2])
    assert not overlaps(intervals[0], intervals[2])
    assert not starts(intervals[3], intervals[4])


def test_compositions():
    elems = [tuple([1, 2]), tuple([4, 6]), tuple([8, 10])]

    rel = composition([[before, before]], elems)
    assert apply_composition(rel, [[tuple([1, 2]), tuple([8, 10])]])

    rel = composition([[before, before]], elems)
    assert not apply_composition(rel, [[tuple([4, 6]), tuple([8, 10])]])

    elems.append(tuple([1, 5]))
    rel = composition([[starts, before]], elems)
    assert apply_composition(rel, [[tuple([1, 2]), tuple([8, 10])]])

    elems.append(tuple([1, 2]))
    rel = composition([[equals, starts]], elems)
    assert apply_composition(rel, [[tuple([1, 2]), tuple([1, 5])]])

    elems.append(tuple([2, 5]))
    rel = composition([[meets, overlaps]], elems)
    assert apply_composition(rel, [[tuple([1, 2]), tuple([4, 6])]])

    # multiple compositions
    elems.append(tuple([1, 2]))
    elems.append(tuple([1, 2]))
    rel = composition([[equals, equals], [equals, equals]], elems)
    assert apply_composition(
        rel, [[tuple([1, 2]), tuple([1, 2])], [tuple([1, 2]),
                                               tuple([1, 2])]]
    )

    elems.append(tuple([5, 8]))
    elems.append(tuple([0, 1]))
    rel = composition([[before, overlaps], [overlaps, overlaps]], elems)
    assert apply_composition(
        rel, [[tuple([0, 1]), tuple([4, 6])], [tuple([2, 5]),
                                               tuple([5, 8])]]
    )


def test_calculus_axioms():
    elems = [tuple(random.randint(1, 100, size=2)) for _ in range(10)]

    # Huntington's axiom
    r, s = random.choice([
        before, overlaps, during, meets, starts, finishes, equals
    ], 2)
    i, j = random.choice(range(len(elems)), 2, replace=False)
    assert not (not r(elems[i], elems[j]) or not s(elems[i], elems[j])) or (
        not ((not r(elems[i], elems[j])) or s(elems[i], elems[j]))
    ) == r(elems[i], elems[j])

    # identity
    i, j = random.choice(range(len(elems)), 2)
    elems.append(elems[j])
    rel = composition([[meets, equals]], elems)
    assert apply_composition(rel, [[elems[i], elems[j]]
                                   ]) == meets(elems[i], elems[j])

    i, j = random.choice(range(len(elems)), 2)
    elems.append(elems[i])
    rel = composition([[equals, meets]], elems)
    assert apply_composition(rel, [[elems[i], elems[j]]
                                   ]) == meets(elems[i], elems[j])

    # involution
    for op in [before, overlaps, during, meets, starts, finishes, equals]:
        i, j = random.choice(range(len(elems)), 2, replace=False)
        converse(converse(op))(elems[i], elems[j]) == op(elems[i], elems[j])

    # associativity
    r, s, t = random.choice([
        before, overlaps, during, meets, starts, finishes, equals
    ], 3)
    c1 = composition([[r, s]], elems)
    c2 = composition([[s, t]], elems)
    c = composition([[r, s], [s, t]], elems)
    i, j, k, l = random.choice(range(len(elems)), 4, replace=False)
    t1, t2, t3, t4 = elems[i], elems[j], elems[k], elems[l]
    assert (
        apply_composition(c1, [[t1, t2]]) and
        apply_composition(c2, [[t3, t4]])
    ) == apply_composition(c, [[t1, t2], [t3, t4]])

    # distributivity
    r, s, t = random.choice([
        before, overlaps, during, meets, starts, finishes, equals
    ], 3)
    c1 = composition([[r, t]], elems)
    c2 = composition([[s, t]], elems)
    c = composition([[random.choice([s, r]), t]], elems)
    i, j = random.choice(range(len(elems)), 2, replace=False)
    app = apply_composition(c, [[elems[i], elems[j]]])
    assert (apply_composition(c1, [[
        elems[i], elems[j]
    ]]) == app) or (apply_composition(c2, [[elems[i], elems[j]]]) == app)

    # inv-distrib
    r, s = random.choice([
        before, overlaps, during, meets, starts, finishes, equals
    ], 2)
    i, j, k, l = random.choice(range(len(elems)), 4, replace=False)
    [i, j, k,
     l] = [(q, p) for (p, q) in [elems[i], elems[j], elems[k], elems[l]]]
    assert any([r(i, j), s(k,
                           l)]) == any([converse(r)(j, i),
                                        converse(s)(l, k)])

    # inv-involutive-distr
    s, t = random.choice([
        before, overlaps, during, meets, starts, finishes, equals
    ], 2)
    c = composition([[s, t]], elems)
    inv_c = composition([[converse(t), converse(s)]], elems)
    i, j = random.choice(range(len(elems)), 2, replace=False)
    assert apply_composition(c, [[
        elems[i], elems[j]
    ]], conversion=True) == apply_composition(inv_c, [[elems[j], elems[i]]])

    # Tarski/ de Morgan
    r, s = random.choice([
        before, overlaps, during, meets, starts, finishes, equals
    ], 2)
    i, j = random.choice(range(len(elems)), 2, replace=False)
    c = composition([[converse(r), negate(r)]], elems)
    c2 = composition([[r, s]], elems)
    assert (
        apply_composition(
            c, [[elems[i], elems[j]]],
            negate_whole_expression=False,
            negations=[False, True]
        ) and apply_composition(
            c2, [[elems[i], elems[j]]], negate_whole_expression=True
        ) and (not s)
    ) == (not s)


def test_get_interval_relations_of_regions():
    r1 = Region((1, 1, 1), (2, 2, 2))
    r2 = Region((5, 5, 5), (8, 8, 8))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['b', 'b', 'b'])

    r1 = Region((1, 1, 1), (10, 10, 10))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['di', 'di', 'di'])

    r1 = Region((1, 1, 1), (6, 6, 6))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['o', 'o', 'o'])

    r2 = Region((1, 1, 1), (2, 2, 2))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['si', 'si', 'si'])

    r2 = Region((1, 1, 1), (6, 6, 6))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['e', 'e', 'e'])

    r1 = Region((5, 5, 5), (8, 8, 8))
    r2 = Region((8, 7, 12), (10, 8, 14))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['m', 'fi', 'b'])
    assert get_intervals_relations(
        r2.bounding_box.limits, r1.bounding_box.limits
    ) == tuple(['mi', 'f', 'bi'])

    r1 = Region((5, 5, 5), (8, 8, 8))
    r2 = Region((3, 3, 7), (6, 6, 9))
    assert get_intervals_relations(
        r1.bounding_box.limits, r2.bounding_box.limits
    ) == tuple(['oi', 'oi', 'o'])
    assert get_intervals_relations(
        r2.bounding_box.limits, r1.bounding_box.limits
    ) == tuple(['o', 'o', 'oi'])



import operator

import numpy as np
import pytest

from ...datalog import Fact
from ...expressions import Constant, Symbol
from ...logic import Conjunction, Implication, Union
from ...relational_algebra import (
    ColumnStr,
    ExtendedProjection,
    ExtendedProjectionListMember,
    NamedRelationalAlgebraFrozenSet,
    RenameColumn,
    str2columnstr_constant,
)
from ...relational_algebra_provenance import ProvenanceAlgebraSet
from .. import dichotomy_theorem_based_solver, weighted_model_counting
from ..cplogic import testing
from ..cplogic.program import CPLogicProgram
from ..dichotomy_theorem_based_solver import ProbSemiringSolver
from ..exceptions import (
    NotEasilyShatterableError,
    NotHierarchicalQueryException,
)

try:
    from contextlib import nullcontext
except ImportError:
    from contextlib import suppress as nullcontext

EQ = Constant(operator.eq)

ans = Symbol("ans")
P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
H = Symbol("H")
A = Symbol("A")
B = Symbol("B")
C = Symbol("C")
w = Symbol("w")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")
p = Symbol("p")

a = Constant("a")
b = Constant("b")
c = Constant("c")


@pytest.fixture(
    params=((weighted_model_counting, dichotomy_theorem_based_solver)),
    ids=["SDD-WMC", "dichotomy-Safe query"],
)
def solver(request):
    return request.param


def test_deterministic(solver):
    """
    We define the program

        P(x) <- Q(x)

    And we expect the provenance set resulting from the
    marginalisation of P(x) to be

        _p_ | x
        ====|===
        1.0 | a
        1.0 | b

    """
    code = Union(
        (
            Fact(Q(a)),
            Fact(Q(b)),
            Implication(P(x), Q(x)),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    query = Implication(ans(x), P(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(1.0, "a"), (1.0, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_deterministic_conjunction_varying_arity(solver):
    code = Union(
        (
            Fact(Q(a, b)),
            Fact(P(a)),
            Fact(P(b)),
            Implication(Z(x, y), Conjunction((Q(x, y), P(x), P(y)))),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(1.0, "a", "b")], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_simple_bernoulli(solver):
    """
    We define the program

        P(a) : 0.7 <- T
        P(b) : 0.8 <- T

    And expect the provenance set resulting from the
    marginalisation of P(x) to be

        _p_ | x
        ====|===
        0.7 | a
        0.8 | b

    """
    code = Union(())
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    cpl_program.add_probabilistic_facts_from_tuples(
        P, {(0.7, "a"), (0.8, "b")}
    )
    query = Implication(ans(x), P(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(0.7, "a"), (0.8, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_bernoulli_conjunction(solver):
    code = Union((Implication(Z(x), Conjunction((P(x), Q(x), R(x)))),))
    probfacts_sets = {
        P: {(1.0, "a"), (0.5, "b")},
        Q: {(0.9, "b"), (0.1, "c")},
        R: {(0.9, "b"), (0.1, "c")},
    }
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    for pred_symb, pfact_set in probfacts_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    query = Implication(ans(x), Z(x))
    result = solver.solve_succ_query(query, cpl_program)

    expected = testing.make_prov_set([(0.9 * 0.9 * 0.5, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_multi_level_conjunction(solver):
    """
    We consider the program

           P(a) : 0.2 <- T
           Q(a) : 0.9 <- T
        R(a, a) : 0.1 <- T
        R(a, b) : 0.5 <- T
                 Z(x) <- P(x), Q(x)
              H(x, y) <- Z(x), R(x, y)

    And expect the prov set resulting from the
    marginalisation of H(x, y) to be

        _p_   | x  | y
        ======|====|===
        0.018 | a  | a
        0.09  | a  | b

    """
    probfacts_sets = {
        P: {(0.2, "a")},
        Q: {(0.9, "a")},
        R: {(0.1, "a", "a"), (0.5, "a", "b")},
    }
    code = Union(
        (
            Implication(Z(x), Conjunction((P(x), Q(x)))),
            Implication(H(x, y), Conjunction((Z(x), R(x, y)))),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    for pred_symb, pfact_set in probfacts_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    query = Implication(ans(x, y), H(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.2 * 0.9 * 0.1, "a", "a"),
            (0.2 * 0.9 * 0.5, "a", "b"),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_intertwined_conjunctions_and_probfacts(solver):
    """
    We consider the program

        P(a) : 0.8  <-  T
        C(a) : 0.5  <-  T
        C(b) : 0.9  <-  T
              A(x)  <-  B(x), C(x)
              B(x)  <-  P(x)
              Z(x)  <-  A(x), B(x), C(x)

    And expect SUCC[ Z(x) ] to yield the provenance relation

        _p_ | x
        ====|===
        0.4 | a

    """
    cpl_code = Union(
        (
            Implication(A(x), Conjunction((B(x), C(x)))),
            Implication(B(x), P(x)),
            Implication(Z(x), Conjunction((A(x), B(x), C(x)))),
        )
    )
    cpl = CPLogicProgram()
    cpl.walk(cpl_code)
    cpl.add_probabilistic_facts_from_tuples(P, {(0.8, "a")})
    cpl.add_probabilistic_facts_from_tuples(C, {(0.5, "a"), (0.9, "b")})
    query = Implication(ans(y), Z(y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([(0.8 * 0.5, "a")], ("_p_", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_simple_probchoice(solver):
    pchoice_as_sets = {P: {(0.2, "a"), (0.8, "b")}}
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    query = Implication(ans(x), P(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.2, "a"),
            (0.8, "b"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_mutual_exclusivity(solver):
    pchoice_as_sets = {P: {(0.2, "a"), (0.8, "b")}}
    pfact_sets = {Q: {(0.5, "a", "b")}}
    code = Union((Implication(Z(x, y), Conjunction((P(x), P(y), Q(x, y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    qpred = Z(x, y)
    exp, result = testing.inspect_resolution(qpred, cpl_program)
    assert isinstance(exp, RenameColumn)
    assert isinstance(exp.relation, RenameColumn)
    expected = testing.make_prov_set([], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_multiple_probchoices_mutual_exclusivity(solver):
    pchoice_as_sets = {
        P: {(0.2, "a"), (0.8, "b")},
        Q: {(0.5, "a", "b"), (0.4, "b", "c"), (0.1, "b", "b")},
    }
    rule = Implication(Z(x, y), Conjunction((P(x), Q(y, y))))
    code = Union((rule,))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    qpred = Z(x, y)
    exp, result = testing.inspect_resolution(qpred, cpl_program)
    assert isinstance(exp, RenameColumn)
    assert isinstance(exp.relation, RenameColumn)
    expected = testing.make_prov_set(
        [(0.2 * 0.1, "a", "b"), (0.8 * 0.1, "b", "b")], ("_p_", "x", "y")
    )
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.slow
def test_large_probabilistic_choice(solver):
    n = int(10000)
    with testing.temp_seed(42):
        probs = np.random.rand(n)
    probs = probs / probs.sum()
    pchoice_as_sets = {P: {(float(prob), i) for i, prob in enumerate(probs)}}
    pfact_sets = {Q: {(0.5, 0, 0), (0.5, 0, 1)}}
    code = Union((Implication(Z(x, y), Conjunction((P(x), Q(x, y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * probs[0], 0, 0),
            (0.5 * probs[0], 0, 1),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_simple_existential(solver):
    """
    We define the following program

        P(a, a) : 0.2 v P(a, b) : 0.8 <- T
                           Q(x) <- ∃y, P(x, y)

    We expect the following to hold

        - Pr[P(a, a)] = 0.2
        - Pr[P(a, b)] = 0.8
        - Pr[Q(a)] = 1.0

    """
    pchoice_as_sets = {P: {(0.2, "a", "a"), (0.7, "a", "b"), (0.1, "c", "c")}}
    code = Union((Implication(Q(x), P(x, y)),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    exp, result = testing.inspect_resolution(Q(x), cpl_program)
    expected = testing.make_prov_set([(0.9, "a"), (0.1, "c")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_existential_in_conjunction(solver):
    pchoice_as_sets = {
        P: {(0.2, "a", "b", "c"), (0.4, "b", "b", "c"), (0.4, "b", "a", "c")},
        Z: {(0.5, "b"), (0.5, "d")},
    }
    code = Union((Implication(Q(x), Conjunction((Z(y), P(x, y, z)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    exp, result = testing.inspect_resolution(Q(x), cpl_program)
    expected = testing.make_prov_set([(0.1, "a"), (0.2, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_existential_alternative_variables(solver):
    pchoice_as_sets = {
        P: {(0.8, "a", "b"), (0.1, "c", "d"), (0.1, "d", "e")},
    }
    pfact_sets = {
        Z: {(0.2, "a"), (0.7, "e")},
    }
    code = Union(
        (
            Fact(R(Constant[str]("a"))),
            Fact(R(Constant[str]("b"))),
            Implication(H(x), Conjunction((Z(y), P(y, x)))),
        )
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(z), H(z))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(0.2 * 0.8, "b")], ("prob", "z"))
    assert testing.eq_prov_relations(result, expected)


def test_multilevel_existential(solver):
    pchoice_as_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c")},
        R: {(0.1, "a"), (0.4, "b"), (0.5, "c")},
        Q: {(0.9, "a"), (0.1, "c")},
        Z: {(0.1, "b"), (0.9, "c")},
    }
    code = Union(
        (
            Implication(H(x, y), Conjunction((R(x), Z(y)))),
            Implication(A(x), Conjunction((H(x, y), P(y, x)))),
            Implication(B(x), Conjunction((A(x), Q(y)))),
            Implication(C(x), H(x, y)),
        )
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), H(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.1 * 0.1, "a", "b"),
            (0.1 * 0.9, "a", "c"),
            (0.4 * 0.1, "b", "b"),
            (0.4 * 0.9, "b", "c"),
            (0.5 * 0.1, "c", "b"),
            (0.5 * 0.9, "c", "c"),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)

    query = Implication(ans(z), C(z))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.1, "a"),
            (0.4, "b"),
            (0.5, "c"),
        ],
        ("_p_", "z"),
    )
    assert testing.eq_prov_relations(result, expected)

    query = Implication(ans(z), B(z))

    if solver is dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [(0.5 * 0.1 * 0.5, "c")],
            ("_p_", "z"),
        )
        assert testing.eq_prov_relations(result, expected)


def test_repeated_antecedent_predicate_symbol(solver):
    """
    We consider the simple program

        P(a) : 0.4  <-  T
        P(b) : 0.7  <-  T
           Q(x, y)  <-  P(x), P(y)

    Possible outcomes are

        { }                     with prob   (1 - 0.4) * (1 - 0.7)
        { P(a), Q(a, a) }       with prob   0.4 * (1 - 0.7)
        { P(b), Q(b, b) }       with prob   0.7 * (1 - 0.4)
        { P(a), P(b),
          Q(a, a), Q(a, b),     with prob   0.4 * 0.7
          Q(b, b), Q(b, a) }

    We expected the following provenance set to result from the
    succ query prob[Q(x, y)]?

        _p_                         | x | y
        ----------------------------|---|---
        0.4 * (1 - 0.7) + 0.4 * 0.7 | a | a
        0.7 * (1 - 0.4) + 0.4 * 0.7 | b | b
        0.4 * 0.7                   | a | b
        0.4 * 0.7                   | b | a

    """
    pfact_sets = {
        P: {(0.4, "a"), (0.7, "b")},
    }
    code = Union((Implication(Q(x, y), Conjunction((P(x), P(y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x, y), Q(x, y))

    if solver is dichotomy_theorem_based_solver:
        context = pytest.raises(NotEasilyShatterableError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.4 * (1 - 0.7) + 0.4 * 0.7, "a", "a"),
                (0.7 * (1 - 0.4) + 0.4 * 0.7, "b", "b"),
                (0.4 * 0.7, "a", "b"),
                (0.4 * 0.7, "b", "a"),
            ],
            ("_p_", "x", "y"),
        )
        assert testing.eq_prov_relations(result, expected)


def test_fake_neurosynth(solver):
    TermInStudy = Symbol("TermInStudy")
    ActivationReported = Symbol("ActivationReported")
    SelectedStudy = Symbol("SelectedStudy")
    TermAssociation = Symbol("TermAssociation")
    Activation = Symbol("Activation")
    s = Symbol("s")
    t = Symbol("t")
    v = Symbol("v")
    code = Union(
        (
            Implication(
                TermAssociation(t),
                Conjunction([TermInStudy(t, s), SelectedStudy(s)]),
            ),
            Implication(
                Activation(v),
                Conjunction([ActivationReported(v, s), SelectedStudy(s)]),
            ),
        )
    )
    pfact_sets = {
        TermInStudy: {
            (0.001, "memory", "1"),
            (0.002, "memory", "2"),
            (0.015, "visual", "2"),
            (0.004, "memory", "3"),
            (0.005, "visual", "4"),
            (0.0001, "memory", "4"),
            (0.01, "visual", "5"),
        },
        ActivationReported: {
            (1.0, "v1", "1"),
            (1.0, "v2", "1"),
            (1.0, "v3", "2"),
            (1.0, "v1", "3"),
            (1.0, "v1", "4"),
        },
    }
    pchoice_as_sets = {
        SelectedStudy: {
            (0.2, "1"),
            (0.2, "2"),
            (0.2, "3"),
            (0.2, "4"),
            (0.2, "5"),
        }
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(t), TermAssociation(t))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (
                sum(t[0] for t in pfact_sets[TermInStudy] if t[1] == term) / 5,
                term,
            )
            for term in ("memory", "visual")
        ],
        ("_p_", "t"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_conjunct_pfact_equantified_pchoice(solver):
    pfact_sets = {P: {(0.8, "a", "s1"), (0.5, "a", "s2"), (0.1, "b", "s2")}}
    pchoice_as_sets = {Z: {(0.6, "s1"), (0.4, "s2")}}
    code = Union((Implication(Q(x), Conjunction((P(x, y), Z(y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.6 * 0.8 + 0.4 * 0.5, "a"),
            (0.1 * 0.4, "b"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_shatterable_query(solver):
    pfact_sets = {P: {(0.8, "a", "1"), (0.5, "a", "2"), (0.1, "b", "2")}}
    code = Union((Implication(Q(x), Conjunction((P(a, x), P(b, x)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * 0.1, "2"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_shatterable_query_2(solver):
    pfact_sets = {
        P: {(0.8, "a", "c", "1"), (0.5, "a", "c", "2"), (0.1, "b", "b", "2")}
    }
    code = Union((Implication(Q(x), Conjunction((P(a, c, x), P(b, b, x)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * 0.1, "2"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_program_with_variable_equality(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    pfact_sets = {
        Q: {(0.2, "a"), (0.3, "b"), (0.4, "c")},
    }
    pchoice_as_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c")},
    }
    code = Union(
        (Implication(Z(x, y), Conjunction((P(y, x), Q(y), EQ(y, a)))),)
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * 0.2, "b", "a"),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_repeated_variable_probabilistic_rule(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q, [(0.2, 7, 7, 2), (0.5, 7, 8, 4)]
    )
    cpl.walk(Implication(H(x, x), Q(x, x, y)))
    query = Implication(ans(x, y), H(x, y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([(0.2, 7, 7)], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_repeated_variable_with_constant_in_head(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 7, 8), (0.6, 8, 9), (0.9, 8, 8)],
    )
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.4, 8), (0.6, 9)],
    )
    cpl.walk(Implication(R(Constant[int](8), x), Conjunction((Q(x, x), P(x)))))
    query = Implication(ans(x, y), R(x, y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([(0.4 * 0.9, 8, 8)], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_empty_result_program(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    rule = Implication(R(Constant(2), Constant(3)), Conjunction((Q(x),)))
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 7)],
    )
    cpl.walk(rule)
    query = Implication(ans(x), R(x, x))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([], ("_p_",))
    assert testing.eq_prov_relations(result, expected)


def test_program_with_probchoice_selfjoin(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.2, "a"), (0.8, "b")],
    )
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.6, "a"), (0.8, "b")],
    )
    rule = Implication(R(x, y), Conjunction((Q(x), P(x), P(y))))
    cpl.walk(rule)
    query = Implication(ans(x, y), R(x, y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.2 * 0.6, "a", "a"), (0.8 * 0.8, "b", "b")], ("_p_", "x", "y")
    )
    assert testing.eq_prov_relations(result, expected)


def test_probchoice_selfjoin_multiple_variables(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.2, "a", "b"), (0.8, "b", "c")],
    )
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.6, "a"), (0.8, "b")],
    )
    rule = Implication(R(x, y, z, w), Conjunction((Q(x), P(x, y), P(z, w))))
    cpl.walk(rule)
    query = Implication(ans(x, y, z, w), R(x, y, z, w))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.2 * 0.6, "a", "b", "a", "b"), (0.8 * 0.8, "b", "c", "b", "c")],
        ("_p_", "x", "y", "z", "w"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_probchoice_selfjoin_multiple_variables_shared_var(solver):
    if solver is not dichotomy_theorem_based_solver:
        pytest.skip()
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.2, "a", "b"), (0.8, "b", "b")],
    )
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.6, "a", "b"), (0.8, "b", "a"), (0.2, "b", "b")],
    )
    rule = Implication(R(x, y, z), Conjunction((Q(x, y), P(x, y), P(z, x))))
    cpl.walk(rule)
    query = Implication(ans(x, y, z), R(x, y, z))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.2 * 0.8, "b", "b", "b")],
        ("_p_", "x", "y", "z"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_probsemiring_extended_proj():
    provset = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2, "a", "b"),
                (0.3, "b", "a"),
                (0.5, "c", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    proj_list = [
        ExtendedProjectionListMember(
            str2columnstr_constant("x"), str2columnstr_constant("x")
        ),
        ExtendedProjectionListMember(
            str2columnstr_constant("y"), str2columnstr_constant("y")
        ),
        ExtendedProjectionListMember(
            Constant("d"), str2columnstr_constant("z")
        ),
    ]
    proj = ExtendedProjection(provset, proj_list)
    solver = ProbSemiringSolver()
    result = solver.walk(proj)
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y", "z"),
            [
                (0.2, "a", "b", "d"),
                (0.3, "b", "a", "d"),
                (0.5, "c", "c", "d"),
            ],
        ),
        ColumnStr("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_probsemiring_forbidden_extended_proj_missing_nonprov_cols():
    provset = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2, "a", "b"),
                (0.3, "b", "a"),
                (0.5, "c", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    proj_list = [
        ExtendedProjectionListMember(
            str2columnstr_constant("x"), str2columnstr_constant("x")
        ),
        ExtendedProjectionListMember(
            Constant("d"), str2columnstr_constant("z")
        ),
    ]
    proj = ExtendedProjection(provset, proj_list)
    solver = ProbSemiringSolver()
    with pytest.raises(ValueError):
        solver.walk(proj)


def test_probsemiring_forbidden_extended_proj_on_provcol():
    provset = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2, "a", "b"),
                (0.3, "b", "a"),
                (0.5, "c", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    proj_list = [
        ExtendedProjectionListMember(
            str2columnstr_constant("x"), str2columnstr_constant("x")
        ),
        ExtendedProjectionListMember(
            str2columnstr_constant("y"), str2columnstr_constant("y")
        ),
        ExtendedProjectionListMember(
            Constant("d"), str2columnstr_constant("_p_")
        ),
    ]
    proj = ExtendedProjection(provset, proj_list)
    solver = ProbSemiringSolver()
    with pytest.raises(ValueError):
        solver.walk(proj)



import pytest

from ...datalog.expressions import Fact
from ...exceptions import ForbiddenRecursivityError, UnsupportedProgramError
from ...expressions import Constant, Symbol
from ...logic import TRUE, Conjunction, Implication, Union
from ..cplogic.program import CPLogicProgram
from ..expressions import PROB, ProbabilisticPredicate, ProbabilisticQuery
from ..stratification import stratify_program

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
S = Symbol("S")
T = Symbol("T")
A = Symbol("A")
B = Symbol("B")
C = Symbol("C")
WLQ = Symbol("WLQ")
WLQ1 = Symbol("WLQ1")
WLQ2 = Symbol("WLQ2")
Query = Symbol("Query")
x = Symbol("x")
y = Symbol("y")
p = Symbol("p")
p1 = Symbol("p1")
p2 = Symbol("p2")
a = Constant("a")
b = Constant("b")
c = Constant("c")


def test_stratify_deterministic():
    code = Union(
        (
            Fact(Q(a)),
            Fact(R(b)),
            Fact(S(a)),
            Implication(P(x, y), Conjunction((Q(x), R(y)))),
            Implication(Z(x), S(x)),
        )
    )
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(y), Conjunction((Z(x), P(x, y))))
    idbs = stratify_program(query, program)
    assert len(idbs["deterministic"].formulas) == 2
    assert "probabilistic" not in idbs
    assert "post_probabilistic" not in idbs


def test_stratify_deterministic_probabilistic():
    code = Union(
        (
            Fact(Q(a)),
            Fact(R(b)),
            Implication(ProbabilisticPredicate(Constant(0.5), S(a)), TRUE),
            Implication(P(x, y), Conjunction((Q(x), R(y)))),
            Implication(Z(x), S(x)),
        )
    )
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(y), Conjunction((Z(x), P(x, y))))
    idbs = stratify_program(query, program)
    assert len(idbs["deterministic"].formulas) == 1
    assert idbs["deterministic"].formulas[0].consequent.functor == P
    assert len(idbs["probabilistic"].formulas) == 1
    assert idbs["probabilistic"].formulas[0].consequent.functor == Z
    assert "post_probabilistic" not in idbs


def test_stratify_deterministic_probabilistic_wlq():
    det_idb = [
        Implication(P(x, y), Conjunction((Q(x), R(y)))),
    ]
    prob_idb = [
        Implication(Z(x), S(x)),
        Implication(
            WLQ(x, y, ProbabilisticQuery(PROB, (x, y))),
            Conjunction((P(x, y), Z(x))),
        ),
    ]
    ppq_det_idb = [
        Implication(B(x, p), Conjunction((A(x), A(y), WLQ(x, y, p)))),
    ]
    code = Union(
        [
            Fact(Q(a)),
            Fact(R(b)),
            Implication(ProbabilisticPredicate(Constant(0.5), S(a)), TRUE),
            Fact(A(b)),
            Fact(C(b)),
        ]
        + det_idb
        + prob_idb
        + ppq_det_idb
    )
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(x, p), Conjunction((B(x, p), C(x))))
    idbs = stratify_program(query, program)
    assert set(idbs["deterministic"].formulas) == set(det_idb)
    assert set(idbs["probabilistic"].formulas) == set(prob_idb)
    assert set(idbs["post_probabilistic"].formulas) == set(ppq_det_idb)


def test_stratify_multiple_wlqs():
    det_idb = [
        Implication(P(x, y), Conjunction((Q(x), R(y)))),
    ]
    prob_idb = [
        Implication(Z(x), Conjunction((S(x), T(x, y)))),
        Implication(
            WLQ1(x, y, ProbabilisticQuery(PROB, (x, y))),
            Conjunction((P(x, y), Z(x))),
        ),
        Implication(
            WLQ2(y, ProbabilisticQuery(PROB, (y,))),
            Conjunction((P(y, y), Z(y))),
        ),
    ]
    ppq_det_idb = [
        Implication(
            B(x, p1, p2),
            Conjunction((A(x), A(y), WLQ1(x, y, p1), WLQ2(y, p2))),
        ),
    ]
    code = Union(
        [
            Fact(Q(a)),
            Fact(R(b)),
            Implication(ProbabilisticPredicate(Constant(0.5), S(a)), TRUE),
            Implication(ProbabilisticPredicate(Constant(0.5), T(a)), TRUE),
            Fact(A(b)),
            Fact(C(b)),
        ]
        + det_idb
        + prob_idb
        + ppq_det_idb
    )
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(x, p1, p2), Conjunction((B(x, p1, p2), C(x))))
    idbs = stratify_program(query, program)
    assert set(idbs["deterministic"].formulas) == set(det_idb)
    assert set(idbs["probabilistic"].formulas) == set(prob_idb)
    assert set(idbs["post_probabilistic"].formulas) == set(ppq_det_idb)


def test_wlq_dependence_on_other_wlq():
    det_idb = [
        Implication(P(x, y), Conjunction((Q(x), R(y)))),
    ]
    prob_idb = [
        Implication(Z(x), Conjunction((S(x), T(x, y)))),
        Implication(
            WLQ1(x, y, ProbabilisticQuery(PROB, (x, y))),
            Conjunction((P(x, y), Z(x))),
        ),
        Implication(
            WLQ2(y, p, ProbabilisticQuery(PROB, (y, p))),
            Conjunction((WLQ1(y, y, p), Z(y))),
        ),
    ]
    code = Union(
        [
            Fact(Q(a)),
            Fact(R(b)),
            Implication(ProbabilisticPredicate(Constant(0.5), S(a)), TRUE),
            Implication(ProbabilisticPredicate(Constant(0.5), T(a)), TRUE),
        ]
        + det_idb
        + prob_idb
    )
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(x, p), Conjunction((WLQ2(x, p), C(x))))
    with pytest.raises(UnsupportedProgramError):
        stratify_program(query, program)


def test_cannot_stratify_recursive_program():
    prob_idb = [
        Implication(B(x), Conjunction((A(x), C(x)))),
        Implication(A(x), B(x)),
    ]
    code = Union(
        [
            Fact(C(a)),
            Fact(C(b)),
        ]
        + prob_idb
    )
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(x), A(x))
    with pytest.raises(ForbiddenRecursivityError):
        stratify_program(query, program)


def test_stratification_multiple_post_probabilistic_rules():
    facts = [Fact(R(a)), Fact(R(b)), Fact(B(a)), Fact(B(c)), Fact(B(b))]
    det_idb = [Implication(A(x, y), Conjunction((B(y), R(x))))]
    pfacts = [
        Implication(ProbabilisticPredicate(Constant(0.2), C(a, b)), TRUE)
    ]
    prob_idb = [
        Implication(
            WLQ(x, y, ProbabilisticQuery(PROB, (x, y))),
            Conjunction((A(x, y), C(x, y))),
        )
    ]
    post_prob_idb = [
        Implication(Z(x, p), Conjunction((WLQ(x, y, p), R(y)))),
        Implication(T(p), Z(x, p)),
    ]
    code = Union(tuple(facts + pfacts + prob_idb + det_idb + post_prob_idb))
    program = CPLogicProgram()
    program.walk(code)
    query = Implication(Query(x), T(x))
    idbs = stratify_program(query, program)
    assert set(idbs["deterministic"].formulas) == set(det_idb)
    assert set(idbs["probabilistic"].formulas) == set(prob_idb)
    assert set(idbs["post_probabilistic"].formulas) == set(post_prob_idb)



"""
Translation from First Order Logic to Horn clauses.

This module defines the HornClause expression class and algorithms to translate
First Order Logic expressions to a union of Horn clauses.  Between the
algorithms here are implemented a translation from FOL sentences to an
equivalent expression in safe range normal form, how to obtain the set of range
restricted variables of an expression, how to check if the expression is safe
range and, if so, how to translate the expression to a union of horn clauses.
Furthermore the resulting horn clauses can be translated to a datalog
expression block to use in a solver. The theory behind this is mostly taken
from the chapter 4 of [1]_.

.. [1] Abiteboul, Hull, and Vianu, Foundations of Databases: The Logical Level.
"""
import operator
from functools import reduce
from typing import Callable

from ..datalog.expressions import Fact
from ..datalog.negation import is_conjunctive_negation
from ..exceptions import NeuroLangException
from ..expression_walker import ChainedWalker, PatternWalker, add_match
from ..expressions import ExpressionBlock
from ..type_system import Unknown, is_leq_informative
from . import (
    FALSE,
    TRUE,
    Conjunction,
    Constant,
    Disjunction,
    ExistentialPredicate,
    FunctionApplication,
    Implication,
    LogicOperator,
    Negation,
    Quantifier,
    Symbol,
    Union
)
from .expression_processing import extract_logic_free_variables
from .transformations import (
    DesambiguateQuantifiedVariables,
    EliminateImplications,
    LogicExpressionWalker,
    MoveNegationsToAtoms,
    RemoveUniversalPredicates
)


class NeuroLangTranslateToHornClauseException(NeuroLangException):
    pass


class HornClause(LogicOperator):
    """
    Expression of the form `P(X) \u2190 (Q(X) \u22C0 ... \u22C0 S(X))`
    """

    def __init__(self, head, body):
        self._validate(head, body)
        self.head = head
        self.body = body
        self._symbols = head._symbols | body._symbols

    def _validate(self, head, body):
        if not self._is_valid_as_head(head):
            raise NeuroLangException(
                f"Head must be a literal or FALSE, {head} given"
            )

        if not self._is_valid_as_body(body):
            raise NeuroLangException(
                f"Body must be a Conjunction of literals, a single"
                " literal or TRUE, {body} given"
            )

    def _is_valid_as_head(self, exp):
        return exp == FALSE or self._is_literal(exp)

    def _is_valid_as_body(self, exp):
        return (
            exp == TRUE
            or self._is_literal(exp)
            or (
                isinstance(exp, Conjunction)
                and all(self._is_literal(l) for l in exp.formulas)
            )
        )

    def _is_literal(self, exp):
        return (
            isinstance(exp, FunctionApplication)
            or isinstance(exp, Symbol)
            or isinstance(exp, Constant)
            or (
                isinstance(exp, Negation)
                and (
                    isinstance(exp.formula, FunctionApplication)
                    or isinstance(exp.formula, Symbol)
                    or isinstance(exp.formula, Constant)
                )
            )
        )

    def __repr__(self):
        r = "HornClause{"
        if self.head is not None:
            r += repr(self.head)
            if self.body is not None:
                r += " :- "
        else:
            r += "?- "

        if self.body != TRUE:
            r += repr(self.body)
        r += ".}"
        return r


class HornFact(HornClause):
    """
    Expression of the form `P(X) \u2190 TRUE`
    """

    def __init__(self, head):
        if head == FALSE:
            raise NeuroLangException("A HornFact can not have FALSE as head.")
        super().__init__(head, TRUE)

    def __repr__(self):
        r = "HornFact{"
        if self.head is not None:
            r += repr(self.head)
        r += ".}"
        return r


class MoveNegationsToAtomsOrExistentialQuantifiers(MoveNegationsToAtoms):
    @add_match(Negation(ExistentialPredicate(..., ...)))
    def negated_existential(self, negation):
        quantifier = negation.formula
        h = quantifier.head
        b = self.walk(quantifier.body)
        exp = Negation(ExistentialPredicate(h, b))
        if b != quantifier.body:
            exp = self.walk(exp)
        return exp


def convert_to_srnf(e):
    """
    Convert an expression to safe range normal form.

    Safe range normal form is an equivalent expression
    but where the there are no implications or universal
    quantifiers and also the negations are only applied
    over atoms or existential quantifiers.
    """
    w = ChainedWalker(
        DesambiguateQuantifiedVariables,
        EliminateImplications,
        RemoveUniversalPredicates,
        MoveNegationsToAtomsOrExistentialQuantifiers,
    )
    return w.walk(e)


def is_safe_range(expression):
    """
    Return true if an expression is safe range.

    This function receives an expression in safe range
    normal form and returns true if all its free variables
    are range restricted.
    """
    try:
        return extract_logic_free_variables(
            expression
        ) == range_restricted_variables(expression)
    except NeuroLangException:
        return False


def range_restricted_variables(e):
    return RangeRestrictedVariables().walk(e)


class RangeRestrictedVariables(LogicExpressionWalker):
    """
    The set of variables which are range restricted in a expression.

    The range restricted variables are defined recursively as follows:
    - For atoms: The variables occurring in the arguments.
    - For negation: The empty set.
    - For existential predicates: The restricted variables of the body without
      the head.
    - For conjunction: The union of the restricted variables of the involved
      expressions.
    - For disjunction: The intersection of the restricted variables of the
      involved expressions.

    Also, if an existentially quantified variable is not restricted in the body
    of its predicate then the expression does not have a restricted range. This
    also counts for existentials inside negations.
    """

    @add_match(FunctionApplication)
    def function(self, exp):
        return set([a for a in exp.args if isinstance(a, Symbol)])

    @add_match(Negation)
    def negation(self, exp):
        self.walk(exp.formula)
        return set()

    @add_match(Disjunction)
    def disjunction(self, exp):
        return set.intersection(*self.walk(exp.formulas))

    @add_match(Conjunction)
    def conjunction(self, exp):
        return set.union(*self.walk(exp.formulas))

    @add_match(ExistentialPredicate)
    def existential(self, exp):
        r = self.walk(exp.body)
        if exp.head in r:
            return r - {exp.head}
        # This better could be a return value
        raise NeuroLangException(
            "Some quantified variable is not range restricted"
        )


def convert_srnf_to_horn_clauses(head, expression):
    """
    Converts a safe range query into an union of horn clauses.

    Given a query represented by an _answer_ head and a range restricted
    expression in safe range normal form, returns an union of horn clauses in
    which the result set given to head is equivalent. Also, it is required that
    all the variables appear free in the body.

    The algorithm is implemented using a stack of queries to process, which
    starts with the query given as parameter. In each iteration, a query is
    processed and the result is a Horn clause with a remainder of auxiliar
    queries which appear in the clause. The clause is then added to the result
    and the remainder queries are added to the stack.

    The cases where auxiliar queries are needed are the negation over an
    existential predicate and the disjunction. For the disjunction an auxiliar
    query is introduced for each formula in the disjunction, but all with the
    same head. For a negated existential another query is added because
    negation can only be applied over atoms.

    Care must be taken to ensure that the auxiliar queries are also range
    restricted. To do so, a set of the positive atoms which appear in the
    parent expressions is added alongside each query in the stack. Those
    atoms are precisely the ones that restrict the range of the variables. If
    is the case that a query is not range restricted then some of those atoms
    are added to the body of the query to ensure the range restriction. This
    restricts the result set of the auxiliar queries but does not alter the
    equivalence for the overall query because the range of those variables was
    already restricted in some of its parents.
    """

    if not is_safe_range(expression):
        raise NeuroLangTranslateToHornClauseException(
            "Expression is not safe range: {}".format(expression)
        )
    if not set(head.args) <= range_restricted_variables(expression):
        raise NeuroLangTranslateToHornClauseException(
            "Variables in head ({}) must be present in body ({})".format(
                head, expression
            )
        )

    stack = [(head, expression, set())]
    processed = []

    while stack:
        head, exp, restrictive_atoms = stack.pop()
        body, remainder = ConvertSRNFToHornClause().walk(exp)
        body = _restrict_variables(head, body, restrictive_atoms)
        restrictive_atoms |= _restrictive_atoms(body)
        remainder = [r + (restrictive_atoms,) for r in remainder]
        processed.append(_to_horn_clause(head, body))
        stack += remainder

    return Union(tuple(reversed(processed)))


def _to_horn_clause(head, body):
    if len(body) == 0:
        r = HornFact(head)
    elif len(body) == 1:
        r = HornClause(head, body[0])
    else:
        r = HornClause(head, Conjunction(body))
    return r


def _restrict_variables(head, body, restrictive_atoms):
    while not set(head.args).issubset(_restricted_variables(body)):
        uv = set(head.args) - _restricted_variables(body)
        new_atoms = _choose_restriction_atoms(uv, restrictive_atoms, head)
        body = new_atoms + body
    return body


def _choose_restriction_atoms(unrestricted_variables, available_atoms, head):
    x = list(unrestricted_variables)[0]
    valid_choices = [
        (a, (set(a.args) - set(head.args)))
        for a in available_atoms
        if (x in a.args) and (a != head)
    ]
    valid_choices = sorted(valid_choices, key=lambda t: t[1])
    a, _variables_not_in_head = valid_choices[0]
    return (a,)


def _restricted_variables(body):
    r = set()
    for a in _restrictive_atoms(body):
        r |= _atom_variables(a)
    return r


def _restrictive_atoms(atoms):
    return set(a for a in atoms if _is_restriction(a))


def _is_restriction(atom):
    if isinstance(atom, FunctionApplication):
        if is_leq_informative(atom.functor.type, Unknown):
            return True
        elif atom.functor == operator.eq and any(
            isinstance(a, Constant) for a in atom.args
        ):
            return True
        return not is_leq_informative(atom.functor.type, Callable)
    return False


def _atom_variables(atom):
    return set(s for s in atom.args if isinstance(s, Symbol))


class ConvertSRNFToHornClause(PatternWalker):
    """
    Converts a expression in safe range normal form into the atoms of a horn
    clause and a list of auxiliar queries remaining to be processed.

    See `convert_srnf_to_horn_clauses`.
    """

    @add_match(Conjunction)
    def conjunction(self, exp):
        bodies, remainders = zip(*map(self.walk, exp.formulas))
        return (
            tuple(
                sorted(reduce(operator.add, bodies), key=self._negation_order)
            ),
            reduce(operator.add, remainders),
        )

    def _negation_order(self, exp):
        return 1 if isinstance(exp, Negation) else 0

    @add_match(Disjunction)
    def disjunction(self, exp):
        nh = self._new_head_for(exp)
        return (nh,), [(nh, f) for f in exp.formulas]

    @add_match(ExistentialPredicate)
    def existential(self, exp):
        return self.walk(exp.body)

    @add_match(Negation(FunctionApplication))
    def negated_atom(self, exp):
        return (exp,), []

    @add_match(Negation(ExistentialPredicate))
    def negated_existential(self, exp):
        nh = self._new_head_for(exp.formula)
        return (Negation(nh),), [(nh, exp.formula)]

    @add_match(FunctionApplication)
    def atom(self, exp):
        return (exp,), []

    @add_match(...)
    def unknown(self, exp):
        raise NeuroLangTranslateToHornClauseException(
            "Expression not in safe range normal form: {}".format(exp)
        )

    def _new_head_for(self, exp):
        fv = extract_logic_free_variables(exp)
        S = Symbol.fresh()
        return S(*tuple(fv))


def translate_horn_clauses_to_datalog(horn_clauses):
    """
    Straightforward translation from Horn clauses a datalog expression block.
    """
    return TranslateHornClausesToDatalog().walk(horn_clauses)


class TranslateHornClausesToDatalog(LogicExpressionWalker):
    @add_match(Union)
    def union(self, exp):
        return ExpressionBlock(self.walk(exp.formulas))

    @add_match(Quantifier)
    def quantifier(self, exp):
        return self.walk(exp.body)

    @add_match(HornFact)
    def horn_fact(self, exp):
        return Fact(exp.head)

    @add_match(HornClause)
    def horn_rule(self, exp):
        return Implication(exp.head, self.walk(exp.body))


def fol_query_to_datalog_program(head, exp):
    """
    Returns a datalog program for a given query in first order logic.

    Given a query represented by an _answer_ head and an expression in first
    order logic, converts the expression to safe range normal form and then, if
    the query is safe range, returns an ExpressionBlock with the equivalent
    program in datalog. Throw a NeuroLangTranslateToHornClauseException
    otherwise.
    """
    exp = convert_to_srnf(exp)
    horn_clauses = convert_srnf_to_horn_clauses(head, exp)
    program = translate_horn_clauses_to_datalog(horn_clauses)
    return program


class Fol2DatalogTranslationException(NeuroLangException):
    pass


class Fol2DatalogMixin(PatternWalker):
    @add_match(
        Implication, lambda imp: not is_conjunctive_negation(imp.antecedent)
    )
    def translate_implication(self, imp):
        try:
            program = fol_query_to_datalog_program(
                imp.consequent, imp.antecedent
            )
        except NeuroLangException as e:
            raise Fol2DatalogTranslationException from e
        return self.walk(program)



import collections
import itertools
import operator
from typing import AbstractSet

from ..datalog.expression_processing import enforce_conjunctive_antecedent
from ..expression_pattern_matching import add_match
from ..expression_walker import ExpressionWalker, ReplaceExpressionWalker
from ..expressions import Constant, FunctionApplication, Symbol
from .exceptions import NotEasilyShatterableError
from .probabilistic_ra_utils import ProbabilisticFactSet

EQ = Constant(operator.eq)


def terms_differ_by_constant_term(terms_a, terms_b):
    return any(
        isinstance(term_a, Constant)
        and isinstance(term_b, Constant)
        and term_a != term_b
        for term_a, term_b in zip(terms_a, terms_b)
    )


def all_terms_differ_by_constant_term(list_of_tuple_of_terms):
    return all(
        terms_differ_by_constant_term(terms_a, terms_b)
        for terms_a, terms_b in itertools.combinations(
            list_of_tuple_of_terms, 2
        )
    )


def constant_terms_are_constant_in_all_tuples(list_of_tuple_of_terms):
    arity = len(list_of_tuple_of_terms[0])
    for i in range(arity):
        nb_of_constant_terms = sum(
            isinstance(terms[i], Constant) for terms in list_of_tuple_of_terms
        )
        if 0 < nb_of_constant_terms < len(list_of_tuple_of_terms):
            return False
    return True


def is_easily_shatterable_self_join(list_of_tuple_of_terms):
    """
    A self-join of `m` predicates is easily shatterable if the following two
    conditions are met.

    Firstly, if the `i`th term of one of the predicates is a constant, then the
    `i`th terms of all the other self-joined predicates must also be constants.
    For example, the self-join `P(x, a), P(x, y)` does not meet this condition
    because the second term appears both as a constant (in the first predicate)
    and as a variable (in the second predicate).

    Secondly, all predicates in the self-join must differ by at least one
    constant term. In other words, for two predicates P1 and P2 in the
    self-join, there must be a constant term at some position `i` in P1 whose
    value is different than the constant `i`th term in P2 (we know it's
    constant from the first condition). For example, the self-join `P(x, a),
    P(y, a)` does not meet this condition because the only constant term is `a`
    in both predicates. However, the self-join `P(x, a, b), P(x, a, c)` does
    meet this condition because the predicates differ in their constant value
    of the third term (`b` for the first predicate and `c` for the second
    predicate).

    """
    return constant_terms_are_constant_in_all_tuples(
        list_of_tuple_of_terms
    ) and all_terms_differ_by_constant_term(list_of_tuple_of_terms)


class Shatter(FunctionApplication):
    pass


class QueryEasyShatteringTagger(ExpressionWalker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_args = collections.defaultdict(set)

    @add_match(
        FunctionApplication(ProbabilisticFactSet, ...),
        lambda fa: not isinstance(fa, Shatter)
        and any(isinstance(arg, Constant) for arg in fa.args),
    )
    def shatter_probfact_predicates(self, function_application):
        self._check_can_shatter(function_application)
        self._cached_args[function_application.functor.relation].add(
            function_application.args
        )
        return Shatter(*function_application.unapply())

    @add_match(
        FunctionApplication(ProbabilisticFactSet, ...),
        lambda fa: not isinstance(fa, Shatter),
    )
    def cache_non_constant_args(self, function_application):
        self._check_can_shatter(function_application)
        self._cached_args[function_application.functor.relation].add(
            function_application.args
        )
        return function_application

    def _check_can_shatter(self, function_application):
        pred_symb = function_application.functor.relation
        args = function_application.args
        list_of_tuple_of_terms = list(
            self._cached_args.get(pred_symb, set()).union({args})
        )
        if not is_easily_shatterable_self_join(list_of_tuple_of_terms):
            raise NotEasilyShatterableError(
                f"Cannot easily shatter {pred_symb}-predicates"
            )


class EasyQueryShatterer(ExpressionWalker):
    def __init__(self, symbol_table, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.symbol_table = symbol_table

    @add_match(Shatter(ProbabilisticFactSet, ...))
    def easy_shatter_probfact(self, shatter):
        const_idxs = list(
            i
            for i, arg in enumerate(shatter.args)
            if isinstance(arg, Constant)
        )
        new_relation = self.symbol_table[shatter.functor.relation].value
        new_relation = new_relation.selection(
            {
                new_relation.columns[i + 1]: shatter.args[i].value
                for i in const_idxs
            }
        )
        non_prob_columns = tuple(
            c
            for c in new_relation.columns
            if c != shatter.functor.probability_column.value
        )
        proj_cols = (shatter.functor.probability_column.value,) + tuple(
            non_prob_columns[i]
            for i, arg in enumerate(shatter.args)
            if not isinstance(arg, Constant)
        )
        new_relation = new_relation.projection(*proj_cols)
        new_pred_symb = Symbol.fresh()
        self.symbol_table[new_pred_symb] = Constant[AbstractSet](new_relation)
        non_const_args = tuple(
            arg for arg in shatter.args if not isinstance(arg, Constant)
        )
        new_tagged = ProbabilisticFactSet(
            new_pred_symb, shatter.functor.probability_column
        )
        return FunctionApplication(new_tagged, non_const_args)


class EasyProbfactShatterer(
    QueryEasyShatteringTagger,
    EasyQueryShatterer,
):
    def __init__(self, symbol_table):
        super().__init__(symbol_table)


def query_to_tagged_set_representation(query, symbol_table):
    query = enforce_conjunctive_antecedent(query)
    new_antecedent = ReplaceExpressionWalker(symbol_table).walk(
        query.antecedent
    )
    return query.apply(query.consequent, new_antecedent)


def _check_shatter_fully_solved(shattered_query):
    if any(
        isinstance(formula, Shatter)
        for formula in shattered_query.antecedent.formulas
    ):
        raise NotEasilyShatterableError("Cannot easily shatter query")


def shatter_easy_probfacts(query, symbol_table):
    """
    Remove constants occurring in a given query, possibly removing self-joins.

    A query containing self-joins can be "easily" shattered whenever the
    predicates in the self-joins do not have more than one variable occurring
    in the same term in multiple predicates (e.g. `P(x), P(y)`, both `x` and
    `y` occurr in the same term in both predicates) or the same constant
    occurring in the same term in multiple predicates (e.g. `P(a, x), P(a, b)`,
    `a` occurrs in the same term in both predicates).

    If there is a self-join, the self-joined relation is split into multiple
    relations. These relations are added in-place to the symbol table. The
    returned equivalent query makes use of these relations.

    Parameters
    ----------
    query : Implication
        A conjunctive query (the body can be a single predicate).
    symbol_table : mapping, mutable
        Symbol table containing relations associated with the query's
        relational symbols. This `symbol_table` can be modified in-place by
        this function to add newly generated relational symbols and their
        associated relations.

    Returns
    -------
    Implication
        An equivalent conjunctive query without constants.

    """
    query = enforce_conjunctive_antecedent(query)
    tagged_query = query_to_tagged_set_representation(query, symbol_table)
    shatterer = EasyProbfactShatterer(symbol_table)
    shattered_query = shatterer.walk(tagged_query)
    _check_shatter_fully_solved(shattered_query)
    return shattered_query



import collections
import logging
import os

import numpy as np
import pandas as pd
from pkg_resources import resource_exists, resource_filename

from ..regions import region_set_from_masked_data

try:
    import neurosynth as ns
except ModuleNotFoundError:
    raise ImportError("Neurosynth not installed in the system")


class StudyID(str):
    pass


class TfIDf(float):
    pass


class NeuroSynthHandler(object):
    """
    Class for the management of data provided by neurosynth.
    """

    def __init__(self, ns_dataset=None):
        self._dataset = ns_dataset

    def ns_study_id_set_from_term(self, terms, frequency_threshold=0.05):
        study_ids = self.dataset.get_studies(
            features=terms, frequency_threshold=frequency_threshold
        )
        return set(StudyID(study_id) for study_id in study_ids)

    def ns_study_tfidf_feature_for_terms(self, terms):
        feature_table = self.dataset.feature_table.data
        result_set = set()
        for term in terms:
            if term not in feature_table.columns:
                continue
            result_set |= set(
                (StudyID(tupl[0]), term, tupl[1])
                for tupl in feature_table[[term]].itertuples(
                    index=True, name=None
                )
            )
        return result_set

    def ns_load_dataset(self):

        if resource_exists(
            "neurolang.frontend", "neurosynth_data/dataset.pkl"
        ):
            file = resource_filename(
                "neurolang.frontend", "neurosynth_data/dataset.pkl"
            )
            dataset = ns.Dataset.load(file)
        else:
            path = resource_filename("neurolang.frontend", "neurosynth_data")
            logging.info(
                f"Downloading neurosynth database"
                f" and features in path: {path}"
            )
            dataset = self.download_ns_dataset(path)

        return dataset

    def ns_term_study_associations(self, threshold=1e-3, study_ids=None):
        """
        Load a 2d numpy array containing association between terms and studies
        based on thresholded tf-idf features in the database.

        """
        features = self.dataset.feature_table.data
        terms = features.columns
        if study_ids is None:
            study_ids = features.index.to_series().apply(StudyID)
        study_ids_as_int = study_ids.apply(int)
        features = features.loc[study_ids_as_int]
        features["pmid"] = study_ids
        return (
            features.melt(
                id_vars="pmid",
                var_name="term",
                value_vars=terms,
                value_name="tfidf",
            )
            .query(f"tfidf > {threshold}")[["pmid", "term"]]
            .values
        )

    def ns_reported_activations(self):
        """
        Load a 2d numpy array containing each reported activation in the
        database.

        """
        image_table = self.dataset.image_table
        vox_ids, study_ids_ix = image_table.data.nonzero()
        study_ids = (
            pd.Series(image_table.ids).apply(StudyID).iloc[study_ids_ix]
        )
        return np.transpose([study_ids, vox_ids])

    def ns_study_ids(self):
        return np.expand_dims(
            self.dataset.feature_table.data.index.astype(StudyID), axis=1
        )

    @staticmethod
    def download_ns_dataset(path):
        if not os.path.exists(path):
            os.makedirs(path)
        ns.dataset.download(path=path, unpack=True)
        dataset = ns.Dataset(os.path.join(path, "database.txt"))
        dataset.add_features(os.path.join(path, "features.txt"))
        dataset.save(os.path.join(path, "dataset.pkl"))
        return dataset

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.ns_load_dataset()
        return self._dataset
